// Solace -- Sol Anachronistic Computer Emulation
// A Win32 emulator for the Sol-20 computer.
//
// Copyright (c) Jim Battle, 2000

// this code supplies a very simple scheduler that operates in
// simulated processor time.
//
// a fixed number of timers are supported; timers are one-shot,
// not oscillators.
//
// the timer is only a 32 bit integer, but only 31 bits should be
// used as things might get squirrely near 2^32.  At the nominal
// 2.05 MHz frequency, this gives over 15 minutes of timer for
// a timer event.  Longer timeouts can not be scheduled.
//
// a routine desiring later notification at some specific time
// calls TimerCreate(ticks, *fcn, arg1, arg2), which causes
// 'fcn' to be called with parameters arg1 and arg2 after simulating
// 'ticks' clock cycles.  then event is removed from the queue.

#include "solace.h"


#define NUM_TIMERS 10	// number of active timers allowed

#define TEST_TIMER 0	// if 1, set up a few timers at time 0

static struct {

    int active_timers;		// # of active timers

    int updating;		// semaphore that we are updating

    // rather than updating all N counters every instruction,
    // instead we find the one which will expire first, and
    // transfer that into countdown and startcnt.  as simulated
    // time ticks by, we decrement countdown.
    //
    // when countdown goes <=0, all timers have (startcnt-countdown)
    // ticks removed from their own counter.
    uint32 countdown;
    uint32 startcnt;

    // any timer that has its own counter go <=0 will then
    // cause a callback to the supplied function using the
    // supplied parameters.
    struct {
	uint32    ctr;		// counter until expiration
	timercb_t callback;	// callback function
	uint32    arg1;		// arbitrary value
	uint32    arg2;		// arbitrary value
    } timer[NUM_TIMERS];

} timers;


// set up some events and check that they trigger properly
#if TEST_TIMER
#include <stdio.h>	// for printf
static int g_testtime;
void
TimerTestCB(uint32 arg1, uint32 arg2)
{
    printf("got callback for timer %d after %d clocks\n", arg1, g_testtime);
    if (arg1 == 2)
	(void)TimerCreate(4, TimerTestCB, 5, 0);
}

void
TimerTest(void)
{
    int t1, t2, t3, t4;

    g_testtime = 0;

    t1 = TimerCreate(30, TimerTestCB, 1, 0);
    t2 = TimerCreate(10, TimerTestCB, 2, 0);
    t3 = TimerCreate(50, TimerTestCB, 3, 0);
    t4 = TimerCreate(10, TimerTestCB, 4, 0);
}
#endif


// intialize Timer data structures.  Called once per simulation
void
TimerInit(void)
{
    int i;

    timers.active_timers = 0;
    timers.countdown     = 0;
    timers.updating      = 0;

    for(i=0; i<NUM_TIMERS; i++) {
	timers.timer[i].ctr = 0;	// inactive timer
    }
#if TEST_TIMER
    TimerTest();
#endif
}


// creates a timer.  returns an int, which can be passed back
// to TimerKill to get rid of a timer.
int
TimerCreate(uint32 ticks, timercb_t fcn, uint32 arg1, uint32 arg2)
{
    int tmr;

    // scan through all slots looking for an empty one
    int found = 0;
    for(tmr=0; tmr<NUM_TIMERS; tmr++) {
	if (timers.timer[tmr].ctr == 0) {
	    timers.timer[tmr].ctr      = ticks;
	    timers.timer[tmr].callback = fcn;
	    timers.timer[tmr].arg1     = arg1;
	    timers.timer[tmr].arg2     = arg2;
	    timers.active_timers++;
	    found = 1;
	    break;
	}
    }
    if (!found) {
	WinAlert("Error: ran out of simulated timers");
	exit(-1);
    }

    if (timers.updating) {
	// an expiring timer scheduled a new timer, so at the
	// end of the TimerCredit, the startcnt and countdown
	// get recomputed, and we don't need to tweak them here.
	;
    } else if (timers.active_timers == 1) {
	// this is the only timer
	timers.startcnt = timers.countdown = ticks;
    } else if (ticks < timers.countdown) {
	// the new timer expires before current target
	uint32 diff = (timers.countdown - ticks);
	timers.startcnt  -= diff;
	timers.countdown -= diff;	// == ticks
    } else {
	// the timer state might be that we already have credits lined up
	// for other timers.  rather than doling out the elapsed time to
	// each timer (which may cause some timer to expire, which may cause
	// a callback to try and create another timer, causing reentrancy
	// issues), instead we just add the current credit to the new timer.
	// that is, if N cycles have gone by since the last time we
	// evaluated who to trigger next, we pretend that the new event
	// which is to trigger in M cycles was actually scheduled for M+N
	// ticks N cycles ago.
	int elapsed = (timers.startcnt - timers.countdown);
	timers.timer[tmr].ctr += elapsed;
    }

    return tmr;
}


// kill a timer.  the timer number passed to this function
// is the one returned by the TimerCreate function.
void
TimerKill(int n)
{
    if (timers.timer[n].ctr == 0) {
	WinAlert("Error: killing non-existent simulated timer");
	exit(-1);
    }

    timers.timer[n].ctr = 0;
    timers.active_timers--;
}


// transfer accumulated timer deficit to each active timer.
// n is added to the already elapsed time.
// this shouldn't need to be called very frequently.
// if there are no active timers, startcnt/countdown are both 0.
static void
TimerCredit(uint32 n)
{
    uint32 elapsed;
    int i, flag;

    if ((timers.active_timers == 0) || (n == 0))
	return;	// no timers or no time

    elapsed = n + (timers.startcnt - timers.countdown);
    timers.updating = 1;

    // scan each old timer
    for(i=0; i<NUM_TIMERS; i++) {
	if (timers.timer[i].ctr != 0) {
	    if (timers.timer[i].ctr <= elapsed) {
		// a timer has expired
		timercb_t fcn = timers.timer[i].callback;
		uint32 arg1   = timers.timer[i].arg1;
		uint32 arg2   = timers.timer[i].arg2;
		// destroy timer
		timers.timer[i].ctr = 0;
		timers.active_timers--;
		// call the callback
		fcn(arg1,arg2);
	    } else {
		// a timer hasn't expired
		timers.timer[i].ctr -= elapsed;
	    }
	}
    }


    // determine the new target time
    flag = 0;
    timers.countdown = 0;
    for(i=0; i<NUM_TIMERS; i++) {
	if (timers.timer[i].ctr != 0) {
	    if (!flag || (timers.timer[i].ctr < timers.countdown)) {
		timers.countdown = timers.startcnt = timers.timer[i].ctr;
		flag = 1;
	    }
	}
    }

    timers.updating = 0;	// done updating
}


// each CPU clock corresponds to n base clocks.
// This function changes that scaling.  We support
// only a finite number of scale factors.
#define MAXTICK 100
static int tickscale[MAXTICK];
static int tscale;

void
TimerTickScale(int n)
{
    int i;
    tscale = n;
    for(i=0; i<MAXTICK; i++)
	tickscale[i] = tscale*i;
}


// let 'n' cycles of simulated time go past
void
TimerTick(uint32 n)
{
    // although individual instructions never take this long,
    // it sometimes crops up as a side-effect of hitting breakpoints
    // and such (which is done by effectively advancing time to the
    // end of a timeslice).
    //ASSERT(n<MAXTICK);
    if (n < MAXTICK)
	n = tickscale[n];
    else
	n *= tscale;

#if TEST_TIMER
    g_testtime += n;
    if (g_testtime < 100)
	printf("inc %d, time=%d\n", n, g_testtime);
#endif

#if 0
FIXME
    // here we make one exception -- rather than using the
    // normal scheduler mechanism, we compute horizontal sync
    // for the Sol explicitly because it is triggered very often:
    // on average once every 5-10 instructions.  dirty dirty
    simstate.scanadvctr += n;
    if (simstate.scanadvctr  > simstate.scanadvlimit)
	simstate.scanadvctr -= simstate.scanadvlimit;
#endif

    if (timers.countdown > n) {

	// this should be the most common case
	timers.countdown -= n;

    } else if (timers.countdown > 0) {

	// it is likely that one or more timers has expired.
	// it isn't guaranteed since it is possible that the timer
	// which was nearest to completion was recently TimerKill'd.
	TimerCredit(n);

    } // else no active timers
}

