// Solace -- Sol Anachronistic Computer Emulation
// A Win32 emulator for the Sol-20 computer.
//
// Copyright (c) Jim Battle, 2000, 2001

// this file contains the (supposedly) Win32-independent
// core simulator.  It is responsible for the non-gui
// portions of the emulation.

#include <stdio.h>
#include <string.h>

#include "solace.h"
#include "wintape.h"	// interface to virtual tape drives

#define BRKPT_ADDR(addr) (brkpt_map[(addr)>>5] & (1 << ((addr) & 31)))
#define BRKPT_IO(port)   (brkpt_iomap[(port)])

// scale a floating point time in microseconds to be an
// argument appropriate for the TimerCreate() function.
#define TIMER_US(f) ((int)(14.318180*(f)+0.5))

// ------------------- forward declarations ----------------------

static void WrZ80_norm(word Addr, byte Value);
static byte RdZ80_norm(word Addr);
static byte InZ80_norm(word Port);
static void OutZ80_norm(word Port, byte Value);

static word Exec8080_brkpt(Z80 *R);
static void WrZ80_brkpt(word Addr, byte Value);
static byte RdZ80_brkpt(word Addr);
//static byte OpZ80_brkpt(word Addr);
static byte InZ80_brkpt(word Port);
static void OutZ80_brkpt(word Port, byte Value);

void UpdateTimerInfo(void);
static void handle_scanadv(uint32 arg1, uint32 arg2);
static void handle_kb_stat(uint32 arg1, uint32 arg2);
static void handle_svt_rdtimer(uint32 arg1, uint32 arg2);
static void handle_svt_wrtimer(uint32 arg1, uint32 arg2);
static void report_curstate(void);


// ------------------- state ----------------------

// motherboard state
sysstate_t sysstate;

// simulation state
simstate_t simstate;

// this is the memory image.  parts of it might
// be ghosted by other functions.
uint8 memimage[65536];

// z80 state
Z80 Z80Regs;

// ---------------------------------------
// Logging routines for debugging

#if 0
static char logfilename[256];

static void
OpenLog(char *filename)
{
    FILE *fp;
    strcpy(logfilename, filename);
    fp = fopen(filename, "w");
    fclose(fp);
}

static void
Log(char *str) {
    FILE *fp;
    fp = fopen(logfilename, "a");
    fprintf(fp, str);
    fclose(fp);
}
#endif


// ---------------------------------------

// initialize state that isn't affected by Cold/Warm reset
// it is only ever called once.
void
InitEmu(void)
{
    int i;

    // set CPU speed: 2.04 MHz by default
    Set8080Speed(MHZ_2_04);

    // dipswitch settings
    SetDipSwitch(1, SW1_POLARITY | SW1_SOLIDCUR);	// display properties
    SetDipSwitch(2, 0x00);				// sense switches
    SetDipSwitch(3, 0x20);				// serial port baud rate: 1200 baud default
    SetDipSwitch(4, 0x28);				// serial port control: 8b, no parity, 1 stop default

    // pointers to memory read and write
    // later on we may switch these for debugging
    simstate.exec8080_func = Exec8080;
    simstate.diz80_func = RdZ80_norm;
    simstate.opz80_func = RdZ80_norm;
    simstate.rdz80_func = RdZ80_norm;
    simstate.wrz80_func = WrZ80_norm;

    // reset simulation scheduler
    TimerInit();
    sysstate.scrolltimer  =  0;
    simstate.hscrolltimer = -1;
    sysstate.scanadv      =  0;
    simstate.hscanadv     = -1;
    simstate.hkb_stat     = -1;
    simstate.vt_rdtimer   = -1;
    simstate.vt_wrtimer   = -1;

    // although the Sol is affected by reset (and is handled in SysReset(),
    // the tape player state isn't affected by reset.
    simstate.vt_realtime = 1;
    for(i=0; i<2; i++) {
	sysstate.vt[i].playstate = EMPTY;
	sysstate.vt[i].motor_on  = 0;
	sysstate.vt[i].tape      = NULL;
    }

    // remove all breakpoints
    breakpoint_init();
}


// this routine sets up breakpoint-related crap just before
// resuming execution.
static void
startrun(void)
{
    // generate a bitmap of any active breakpoints
    breakpoint_genmap();

    // we select the fast function if there is no breakpoint
    // of that type, otherwise we pick the slow one that tests
    // for breakpoints on each access.

    if ((simstate.runstate == STEP) || // in case we set tmp breakpoint
	any_breakpoints(BRK_TEMP) ||
	any_breakpoints(BRK_OP) ||
        any_breakpoints(BRK_RD) ||
        any_breakpoints(BRK_WR) ||
        any_breakpoints(BRK_RD16) ||
        any_breakpoints(BRK_WR16) ||
        any_breakpoints(BRK_IN) ||
        any_breakpoints(BRK_OUT)
       )
	simstate.exec8080_func = Exec8080_brkpt;
    else
	simstate.exec8080_func = Exec8080;

    if (any_breakpoints(BRK_OP) || any_breakpoints(BRK_TEMP))
	simstate.opz80_func = RdZ80_norm;	// OpZ80_brkpt; not used
    else
	simstate.opz80_func = RdZ80_norm;

    if (any_breakpoints(BRK_RD) || any_breakpoints(BRK_RD16))
	simstate.rdz80_func = RdZ80_brkpt;
    else
	simstate.rdz80_func = RdZ80_norm;

    if (any_breakpoints(BRK_WR) || any_breakpoints(BRK_WR16))
	simstate.wrz80_func = WrZ80_brkpt;
    else
	simstate.wrz80_func = WrZ80_norm;

    if (any_breakpoints(BRK_IN))
	simstate.inz80_func = InZ80_brkpt;
    else
	simstate.inz80_func = InZ80_norm;

    if (any_breakpoints(BRK_OUT))
	simstate.outz80_func = OutZ80_brkpt;
    else
	simstate.outz80_func = OutZ80_norm;

    // reset debugging state
    Z80Regs.brkpt = 0;
}


void
bkptchange_coresim(void)
{
    startrun();
}


// simulate cold/warm reset
void
SysReset(int warm_reset)
{
    char *msg = "If you see this message, the ROM didn't load";
    unsigned int a;
    int i;

    // init board state
    sysstate.scrollbase  = 0;
    sysstate.scrollshade = 0;
    sysstate.kb_stat = 0;
    sysstate.kb_key  = 0x00;

    if (!warm_reset) {
	// init memory image
	for(a=0; a<65536; a++) {
	    memimage[a] = 0x76;	// HALT instruction
	}
#if 1
	for(a=0xCC00; a<0xCC00+strlen(msg); a++) {
	    memimage[a] = msg[a-0xCC00];
	}
#else
	// initialize with the character set.  useful for debugging
	// the font, although to be most useful, inhibit subsequent
	// ROM load, which will wipe the video display memory clean.
	{
	    int r, c;
	    for(a=0xCC00; a<0xCC00+1024; a++) { memimage[a] = 0x20; }
	    for(r=0; r<16; r++) {
		for(c=0; c<16; c++) {
		    memimage[0xCC00 + r*64+3*c+1] = 16*r+c;
		}
	    }
	}
#endif
    }

    // most virtual tape state is unaffected by reset,
    // as it is off-board
    sysstate.vt_TTBE    = 1;	// tape transmit buffer empty
    sysstate.vt_rxready = 0;	// tape UART state
    sysstate.vt_overrun = 0;	// tape UART state
    sysstate.vt_framing = 0;	// tape UART state
    sysstate.vt_baud    = 1200;	// tape UART control
    for(i=0; i<2; i++)
	sysstate.vt[i].motor_on = 0;	// motor relay

#if 0
    { // for debugging
	int stat;
	stat = TapeFile(1, VTF_LoadTape, "..\\binaries\\foo.svt");
	printf("TapeFile stat=%d\n", stat);
	//stat = vtape_destroy(vtape);
	//printf("vtape destroy stat=%d\n", stat);
    }
#endif

    // reset the processor state
    ResetZ80(&Z80Regs);
    // the Sol memory decoder forces the first four memory
    // ops to read from 0xC0xx instead of 0x00xx.  Rather
    // than complicating the fetcher, we just pretend reset
    // occurs at 0xC000, the address of the personality module.
    Z80Regs.PC.W = 0xC000;

    if (active_debugwin()) {
	if (simstate.runstate != RUNNING)
	    update_debugwin(UPDATE_ALL);
    } else {
	simstate.runstate = RESET;
    }

    simstate.prog_detect = 0;
}


void
LoadDefaultROM(void)
{
    int i;
    for(i=0; i<0x800; i++)
	memimage[0xC000+i] = solos_rom[i];
}


// ------------- set/get various states --------------
// abstract our data structs from user interface code

void
SetDipSwitch(int swtch, byte value)
{
    switch (swtch) {
	case 1: sysstate.dipsw1 = (sw1_t)value; break;
	case 2: sysstate.dipsw2 = value; break;
	case 3: sysstate.dipsw3 = value; break;
	case 4: sysstate.dipsw4 = value; break;
	default:
	    printf("Error: attempt to set dipswitch %d\n", swtch);
    }
}


byte
GetDipSwitch(int swtch)
{
    switch (swtch) {
	case 1: return sysstate.dipsw1;
	case 2: return sysstate.dipsw2;
	case 3: return sysstate.dipsw3;
	case 4: return sysstate.dipsw4;
	default:
	    printf("Error: attempt to get dipswitch %d\n", swtch);
	    return 0x00;
    }
}


void
Set8080Speed(speed_t clkrate)
{
    simstate.up_speed = clkrate;
    UpdateTimerInfo();
}

speed_t
Get8080Speed(void)
{
    return simstate.up_speed;
}


// get play state of indicated unit
void
GetTapeProp(int unit, int prop, int *param)
{
    ASSERT(unit == 0 || unit == 1);
    switch (prop) {
	case TPROP_REALTIME:
	    *param = simstate.vt_realtime;	// ignores unit
	    break;
	case TPROP_PLAYSTATE:
	    *param = sysstate.vt[unit].playstate;
	    break;
	default:
	    (void)vtape_getprop(sysstate.vt[unit].tape, prop, param);
	    break;
    }
}

// set play state of indicated unit
void
SetTapeProp(int unit, int prop, int param)
{
    ASSERT(unit == 0 || unit == 1);
    switch (prop) {
	case TPROP_REALTIME:
	    simstate.vt_realtime = param;	// ignores unit
	    break;
	case TPROP_PLAYSTATE:
	    sysstate.vt[unit].playstate = param;
	    WinTapeNotify(unit, WT_PLAYSTATE, 0);
	    break;
	default:
	    (void)vtape_setprop(sysstate.vt[unit].tape, prop, param);
	    break;
    }
}


// transfer virtual tapes between disk and memory.
// this function does as it is told and does no double
// checking to see if there is a dirty tape or whatever
// already in there.
// returns 0 if OK, !=0 on error.
int
TapeFile(int unit, int action, char *filename)
{
    int stat, pos;

    ASSERT(unit == 0 || unit == 1);

    if (simstate.vt_rdtimer >= 0)
	TimerKill(simstate.vt_rdtimer);
    if (simstate.vt_wrtimer >= 0)
	TimerKill(simstate.vt_wrtimer);

    switch (action) {

	case VTF_NewTape:
	    // get rid of old tape if it exists
	    (void)TapeFile(unit, VTF_EjectTape, NULL);	// slightly recursive! can't fail
	    // actually create new tape
	    sysstate.vt[unit].tape = vtape_create(30);	// 30 minute tape
	    stat = (sysstate.vt[unit].tape == NULL);
	    if (!stat)
		SetTapeProp(unit, TPROP_PLAYSTATE, STOP);
	    break;

	case VTF_LoadTape:
	    stat = TapeFile(unit, VTF_NewTape, NULL);	// slightly recursive!
	    if (stat == 0)
		stat = vtape_fileread(sysstate.vt[unit].tape, filename);
	    if (stat == 0)
		SetTapeProp(unit, TPROP_PLAYSTATE, STOP);
	    break;

	case VTF_SaveTape:
	    SetTapeProp(unit, TPROP_PLAYSTATE, STOP);
	    stat = vtape_filewrite(sysstate.vt[unit].tape, filename);
	    break;

	case VTF_EjectTape:
	    // get rid of old tape if it exists
	    if (sysstate.vt[unit].tape)
		vtape_destroy(sysstate.vt[unit].tape);
	    sysstate.vt[unit].tape = NULL;
	    SetTapeProp(unit, TPROP_PLAYSTATE, EMPTY);
	    stat = 0;
	    break;

	default:
	    ASSERT(0);
    }

    GetTapeProp(unit, TPROP_POSITION, &pos);
    WinTapeNotify(unit, WT_TIME, pos);
    WinTapeNotify(unit, WT_FILESTATE, 0);
    return (stat != 0);
}


// heuristic test to see if a program is already running
int ProgramRunning(void)
{
    return (simstate.prog_detect > 0);
}


// a write to the display ram
void
WrZ80_display(word Addr, byte Value)
{
    int row, col;

    ASSERT((Addr & 0xFC00) == 0xCC00);

    // see if value is unchanged -- eg, writing space on space
    if (Value == memimage[Addr])
	return;

    // update memory image
    memimage[Addr] = Value;

    // Addr = 1100_11rr_rrcc_cccc
    row = (Addr>>6) & 0x0F;
    col = Addr & 0x3F;

    blit_char_to_disp(row,col,Value);
}


// winmain calls this when keyboard input is received
// if we receive two notifications before the first one
// has been processed, the older one is lost.  the kb
// strobe nominally lasts 6 uS, so that if the key is read
// during that window, it will register more than once.
void
SysKeyPress(byte data)
{
    sysstate.kb_stat  = 1;
    if (simstate.hkb_stat < 0)
	simstate.hkb_stat = TimerCreate(TIMER_US(6), handle_kb_stat, 0, 0);
    sysstate.kb_key = data;
}


// when the user interacts with the debugger, the major
// state changes come through here
void
notify_coresim(int event, int param)
{
    switch (event) {
	case DEBUG_KEY:
	    // halt no matter what and pop open debugger
	    simstate.runstate = HALTED;
	    simstate.inactivate_timer = 0;
	    activate_debugwin();
	    report_curstate();
	    break;
	case RUN_KEY:
	    // run if we're halted, halt if we're running
	    if (simstate.runstate == RUNNING) {
		simstate.runstate = HALTED;
		simstate.inactivate_timer = 0;
		activate_debugwin();
		report_curstate();
	    } else {
		simstate.runstate = RUNNING;
		startrun();	// establish normal/debugging fcn ptrs
		simstate.firststep = 1;
		simstate.inactivate_timer = 3;
	    }
	    break;
	case FORCERUN_KEY:
	    // run whether we are currently running or not
	    simstate.runstate = RUNNING;
	    startrun();	// establish normal/debugging fcn ptrs
	    simstate.firststep = 1;
	    simstate.inactivate_timer = 3;
	    break;
	case STEP_KEY:
	    // step over n instructions
	    if (simstate.runstate == RUNNING)
		activate_debugwin();
	    simstate.runstate  = STEP;
	    simstate.stepcount = param;
	    simstate.firststep = 1;
	    startrun();	// establish normal/debugging fcn ptrs
	    simstate.inactivate_timer = 3;
	    break;
	case STEPI_KEY:
	    // step into n instructions
	    if (simstate.runstate == RUNNING)
		activate_debugwin();
	    simstate.runstate  = STEPI;
	    simstate.stepcount = param;
	    simstate.firststep = 1;
	    simstate.inactivate_timer = 3;
	    startrun();	// establish normal/debugging fcn ptrs
	    break;
	case CLOSE_WIN:
	    simstate.runstate = RUNNING;
	    destroy_debugwin();
	    break;
	default:
	    ASSERT(0);
    }
}


// -------------- fulfill emulator hooks ------------

static byte
RdZ80_norm(word Addr)
{
    // note: the SOL adds one wait state on any read in the range of C000-CFFF
    if ((Addr & 0xF000) == 0xC000)
	Z80Regs.ICount--;

    // at this point, we don't check for read side-effects
    // or any memory mapped I/O.  FIXME.
    return(memimage[Addr]);
}


static void
WrZ80_norm(word Addr, byte Value)
{
    // note: the SOL adds one wait state on any write in the range of C000-CFFF
    if ((Addr & 0xF000) == 0xC000)
	Z80Regs.ICount--;

    switch (Addr & 0xFC00) {
    case 0xC000:
    case 0xC400:
	// the ROM -- xC000 to xC7FF -- ignore it
	break;
    case 0xCC00:
	// the display RAM -- xCC00 to xCFFF
	WrZ80_display(Addr, Value);
	break;
    default:
	// simple memory write
	memimage[Addr] = Value;
    }
}


// read a byte from whichever cassette is active, if either one is
// return one of the SVT_* codes.
// it has no sense of time; the caller is responsible for that.
static int
read_cassette(void)
{
    int unit = -1;
    int stat;

    // it would be possible for both to be on, but unusual.
    // we just stick with vt[1] in that case.
    if (sysstate.vt[1].motor_on && (sysstate.vt[1].playstate == PLAY))
	unit = 1;
    else if (sysstate.vt[0].motor_on && (sysstate.vt[0].playstate == PLAY))
	unit = 0;

    if (unit < 0)
	return SVT_NOBYTE;

    stat = vtape_readbyte(sysstate.vt[unit].tape,
			  sysstate.vt_baud,
			  &sysstate.vt_readbyte,
			  &sysstate.vt_overrun,
			  &sysstate.vt_framing);

    switch (stat) {
	case SVT_BYTE:
	    if (sysstate.vt_rxready)
		sysstate.vt_overrun = 1;	// we stacked up two
	    sysstate.vt_rxready = 1;
	    break;
	case SVT_NOBYTE:
	    break;
	case SVT_EOF:
	    SetTapeProp(unit, TPROP_PLAYSTATE, STOP);
	    WinTapeNotify(unit, WT_EOF, 0);
	    break;
	default:
	    break;
    }

    return stat;
}


static byte
InZ80_norm(word Port)
{
    int stat;

    // note: the SOL adds one wait state on any IN operation
    Z80Regs.ICount--;

    switch (Port) {

    case 0xF8: // read serial port status
	// bit 0: serial carrier detect (active low)
	// bit 1: serial data set ready (active low)
	// bit 2: PE: Parity Error--received parity does not compare to
	//            that programmed.
	// bit 3: FE: Framing Error--valid stop bit not received when
	//            expected.
	// bit 4: OE: Overrun Error--CPU did not accept data before it
	//            was replaced with additional data.
	// bit 5: serial clear to send (active low)
	// bit 6: DR: Data Ready--data received by UART is available
	//            when requested.
	// bit 7: TBRE: Transmitter Buffer Register Empty--UART is ready
	//            to accept another word from the Bidirectional
	//            Data Bus.
	// FIXME
	return   (1<<7)	//  TBRE -- for now, swallow everything
	       | (0<<6)	// !DR   -- for now, no input
	       | (0<<5)	//  CTS  (active low)
	       | (0<<4)	// !OE
	       | (0<<3)	// !FE
	       | (0<<2)	// !PE
	       | (0<<1)	//  DSR  (active low)
	       | (0<<0)	//  CD   (active low)
	    ;

    case 0xF9: // read byte from serial port
	return 0x00;	// FIXME

    case 0xFA: // read parallel port, keyboard, and cassette status
	if (!sysstate.vt_rxready && !simstate.vt_realtime)
	    stat = read_cassette();
	return (sysstate.vt_TTBE        << 7)	// cassette TTBE (tape tx buffer empty)
	     | (sysstate.vt_rxready     << 6)	// cassette TDR  (tape data ready)
	     | (0                       << 5)	// unused
	     | (sysstate.vt_overrun     << 4)	// cassette OE   (read overrun error)
	     | (sysstate.vt_framing     << 3)	// cassette FE   (read framing error)
	     | (0                       << 2)	// parallel port device ready (active low)
	     | (1                       << 1)	// parallel port data input ready (active low)
	     | ((~sysstate.kb_stat & 1) << 0)	// KDR (keyboard data ready)
	    ;

    case 0xFB: // read byte from cassette uart
	sysstate.vt_rxready = 0;
	sysstate.vt_overrun = 0;
	sysstate.vt_framing = 0;
	return sysstate.vt_readbyte;

    case 0xFC: // read keyboard
	if (simstate.hkb_stat < 0)
	    sysstate.kb_stat = 0;
	return sysstate.kb_key;

    case 0xFD: // read byte from parallel port and clear !RDY flop
	return 0x00;	// FIXME

    case 0xFE: // return video display scroll timer
	// bit 0: 250 ms timer after each OUT FE (active high)
	// bit 1: horizontal sync signal (active high)
	//        blank 13 chars, video 64 chars, blank & sync 25 chars
	//        == 8.17 us,     == 40.23 us,    == 15.71 us
	//        so active 15.71 us/64.11 us, or 24.5%
	// because hscanadv makes so many transitions, it is scheduled
	// only when it becomes necessary.
	if (simstate.hscanadv < 0)
	    simstate.hscanadv = TimerCreate(100, handle_scanadv, 0, 0);	// start it first time
	return (sysstate.scanadv << 1)
	    |  (sysstate.scrolltimer);

    case 0xFF: // read sense switches -- active low
	return (~GetDipSwitch(2)) & 0xFF;

    default:
	// ignored
#ifdef DEBUG
	printf("Warning: read from unimplemented IO port 0x%02x\n", Port);
#endif
	break;
    }
    return 0x00;
}


// 250 ms after scrolling, turn off scroll timer bit
static void
handle_outfe_timer(uint32 arg1, uint32 arg2)
{
    sysstate.scrolltimer = 0;
}


// simulate a 10 uS or so strobe after keypress
static void
handle_kb_stat(uint32 arg1, uint32 arg2)
{
    simstate.hkb_stat = -1;	// inactive
}


// cassette uart rx channel timer callback.
// this function handles FWD and RWD states too.
// arg1 is the tape unit
static void
handle_svt_rdtimer(uint32 arg1, uint32 arg2)
{
    // one char at  300 baud is 33.333 ms
    // one char at 1200 baud is  8.333 ms
    int time, stat, pos, realtime;
    int unit = arg1;

    simstate.vt_rdtimer = -1;	// inactivate by default
    if (!sysstate.vt[unit].motor_on)
	return;

    GetTapeProp(unit, TPROP_POSITION, &pos);
    GetTapeProp(unit, TPROP_REALTIME, &realtime);

    switch (sysstate.vt[unit].playstate) {

	case EMPTY:
	case STOP:
	case RECORD:
	    // do nothing except disable timer
	    break;

	case PLAY:
	    if (realtime)
		stat = read_cassette();
	    else
		stat = (sysstate.vt_rxready) ? SVT_OK : read_cassette();
	    if (stat != SVT_EOF) {
		// even if we are !realtime, we schedule a callback
		// to make sure that time advances at least as fast
		// as realtime.  this is important because SOLOS goes
		// into a ~1 second idle loop after the cassette motor
		// is first turned on.  we don't want the timer to freeze
		// in this mode because we aren't polling the status reg.
		time = (sysstate.vt_baud == 1200) ? 8333 : 33333;
		simstate.vt_rdtimer =
		    TimerCreate(TIMER_US(time), handle_svt_rdtimer, unit, arg2);
	    }
	    GetTapeProp(unit, TPROP_POSITION, &pos);
	    WinTapeNotify(unit, WT_TIME, pos);
	    break;

	case FWD:
	    // every 10th of a second, adjust pointer.  if not realtime, go 10x faster.
	    pos++;
	    if (vtape_setprop(sysstate.vt[unit].tape, VTPROP_POSITION, pos) == SVT_EOF) {
		// hit end of tape
		SetTapeProp(unit, TPROP_PLAYSTATE, STOP);
		WinTapeNotify(unit, WT_EOF, 0);
	    } else {
		time = (realtime) ? 10000 : 1000;
		simstate.vt_rdtimer =
		    TimerCreate(TIMER_US(time), handle_svt_rdtimer, unit, arg2);
		WinTapeNotify(unit, WT_TIME, pos);
	    }
	    break;

	case REW:
	    // every 10th of a second, adjust pointer
	    pos--;
	    if (pos <= 0) {
		// hit start of tape
		pos = 0;
		SetTapeProp(unit, TPROP_POSITION, pos);
		SetTapeProp(unit, TPROP_PLAYSTATE, STOP);
	    } else {
		time = (realtime) ? 10000 : 1000;
		SetTapeProp(unit, TPROP_POSITION, pos);
		simstate.vt_rdtimer =
		    TimerCreate(TIMER_US(time), handle_svt_rdtimer, unit, arg2);
	    }
	    WinTapeNotify(unit, WT_TIME, pos);
	    break;

	default:
	    ASSERT(0);

    } // switch
}


// cassette uart tx channel timer callback.
//
// if the tape TX UART channel has data ready, write that, otherwise
// just record carrier tone to the tape at the current baud rate.
//
// it is theoretically possible that both tape players are in the
// RECORD state and that both have their motor_on set (although
// SOLOS doesn't do this).
static void
handle_svt_wrtimer(uint32 arg1, uint32 arg2)
{
    // one char at  300 baud is 33.333 ms
    // one char at 1200 baud is  8.333 ms
    int time = (sysstate.vt_baud == 1200) ? 8333 : 33333;
    int unit, stat, pos;
    int data;		// byte value to save
    int carrierflag;	// -1 if carrier instead of data
    int either = 0;	// flag if either unit is actively recording

    simstate.vt_wrtimer = -1;	// inactivate by default

    for(unit=0; unit<2; unit++) {

	if (!sysstate.vt[unit].motor_on)
	    continue;
	if (sysstate.vt[unit].playstate != RECORD)
	    continue;

	if (sysstate.vt_TTBE) {
	    data = 0x00;
	    carrierflag = -1;
	} else {
	    data = sysstate.vt_writebyte;
	    carrierflag = SVT_OK;
	}

	stat = vtape_writebyte(sysstate.vt[unit].tape,
			       sysstate.vt_baud,
			       data, carrierflag);

	if (stat == SVT_EOF) {
	    SetTapeProp(unit, TPROP_PLAYSTATE, STOP);
	    WinTapeNotify(unit, WT_EOF, 0);
	} else {
	    GetTapeProp(unit, TPROP_POSITION, &pos);
	    WinTapeNotify(unit, WT_TIME, pos);
	    either = 1;
	}

    } // for(unit)

    sysstate.vt_TTBE = 1;	// UART TX buffer is now empty

    // schedule next deadline to record carrier.
    // this timer might be killed later if !realtime and a new
    // byte comes in to be written.
    if (either)
	simstate.vt_wrtimer =
	    TimerCreate(TIMER_US(time), handle_svt_wrtimer, 0, 0);
}


/*
   Bit 1 of port 0xFE contains the active high signal "scan adv"
   (scanline advance).  It is related to, but not the same as,
   horizontal blanking.  The waveforms look like this, with the
   numbers below specifying time in characters:


      vid       ________--------------------________________

                   13            64                25


      scan adv  ____________________________----------------_

                            13+64                  25

   each char time is 14.31 MHz * 9 dots, or 628.93 ns.
*/
static void
handle_scanadv(uint32 arg1, uint32 arg2)
{
    int ticks;
    sysstate.scanadv = !sysstate.scanadv;
    ticks = (sysstate.scanadv) ? 25*9 : 77*9;
    simstate.hscanadv = TimerCreate(ticks, handle_scanadv, 0, 0);
}


static void
OutZ80_norm(word Port, byte Value)
{
    // note: the SOL adds one wait state on any OUT operation
    Z80Regs.ICount--;

    switch (Port) {

    case 0xF8: // set serial port RTS
	// bit 4: serial port RTS (ready to send) (active high)
	// FIXME
	break;

    case 0xF9: // write byte to serial port
	// FIXME
	break;

    case 0xFA: // write parallel port and cassette data control bits
	// bit 0: (unused)
	// bit 1: (unused)
	// bit 2: (unused)
	// bit 3: PIE (Parallel Input Enable, active low, resets to 1)
	// bit 4: PUS (Parallel Unit Select, active high, resets to 0)
	// bit 5: tape speed (0=1200 baud, 1=300 baud, resets to 0)
	// bit 6: tape 2 motor enable (active high acc'd to SOLOS, low acc'd to schematic, resets to 0)
	// bit 7: tape 1 motor enable (active high acc'd to SOLOS, low acc'd to schematic, resets to 0)
	// FIXME
	sysstate.vt_baud        = (Value & 0x20) ? 300 : 1200;
	sysstate.vt[0].motor_on = (Value >> 6) & 1;
	sysstate.vt[1].motor_on = (Value >> 7) & 1;
	if (sysstate.vt[0].motor_on && sysstate.vt[1].motor_on) {
	    // FIXME: don't just warn about it
	    static flag = 1;
	    if (flag) {
		flag = 0;
		WinAlert("Warning: both cassette motors enabled -- emulator might not model this properly");
	    }
	}
	break;

    case 0xFB: // write byte to cassette uart
	// see if the uart is still busy writing the previous byte
	if (!sysstate.vt_TTBE) {
	    static flag = 1;
	    if (flag) {
		char msg[80];
		sprintf(msg, "Warning: program at $%04X wrote to cassette data out port while it was busy", Z80Regs.PC.W);
		WinAlert(msg);
		flag = 0;
	    }
	}

	// write it to tape
	sysstate.vt_TTBE      = 0;	// TX busy
	sysstate.vt_writebyte = Value;

	if (simstate.vt_realtime) {
	    int time = (sysstate.vt_baud == 1200) ? 8333 : 33333;
	    if (simstate.vt_wrtimer < 0)
		simstate.vt_wrtimer =
		    TimerCreate(TIMER_US(time), handle_svt_wrtimer, 0, 0);
	} else {
	    // if we're waiting to record carrier, kill it
	    if (simstate.vt_wrtimer >= 0)
		TimerKill(simstate.vt_wrtimer);
	    handle_svt_wrtimer(0, 0);	// write it directly
	}
	break;

    case 0xFD: // write byte to parallel port
	// FIXME
	break;

    case 0xFE: // set display scrollbase
	{
	    if (sysstate.scrolltimer == 1)
		TimerKill(simstate.hscrolltimer);	// one is already active
	    sysstate.scrolltimer = 1;
	    simstate.hscrolltimer =
		    TimerCreate(TIMER_US(250000), handle_outfe_timer, 0, 0);
	}
	sysstate.scrollbase  = ((Value >> 0) & 0xF);
	sysstate.scrollshade = ((Value >> 4) & 0xF);
	WinUpdateScrollBase(sysstate.scrollshade, sysstate.scrollbase);
	break;

    case 0xFF: // optional
	// dudley henderson had a tricked-up Sol for editing
	// tibetan text.  He modified the character generation
	// logic to choose between one of two fonts, toggling
	// between them every time OUT 0xFF was performed.
	//
	// this is a bit impure, but we let the winmain code
	// handle all the display generation logic.
	if (WinOutFFTickle())
	    return;
	// fall through to default error message ...

    default:
	// ignored
#ifdef DEBUG
	printf("Warning: write to unimplemented IO port 0x%02x\n", Port);
#endif
	break;
    }
}


static word
Exec8080_brkpt(Z80 *R)
{
    int addr = R->PC.W;
    word retval;

    if (BRKPT_ADDR(addr) && !simstate.firststep) {

	// potentially a hit -- do slow check
	int hit = breakpoint_test(BRK_OP, R->PC.W, (word)0, &R->brknum);

	if (hit) {
	    R->brkpt = 1;
	    R->reason = BRKPT_BRK;
	    R->XCount = R->ICount;
	    R->ICount = 0;	// terminates execution loop (if any)
	    simstate.firststep = 0;
	    return R->PC.W;
	}

    } // if (BRKPT_ADDR())

    // simulate the instruction
    retval = Exec8080(R);

    if (R->brkpt) {
	// validate breakpoint (rd16 and wr16 are troublesome)
	int hit = breakpoint_verify(&R->brknum);
	if (!hit) {
	    // oops, false alarm
	    R->brkpt  = FALSE;
	    R->ICount = R->XCount;	// restore remaining count
	}
    }

    simstate.firststep = 0;	// clear it in case this is the first time
    return retval;
}


#if 0
// This is not used because the PC breakpoints are detected before
// calling the Exec8080() routine, where this would be called.
// This is done because otherwise we'd have to have breakpoint
// checking in Exec8080() which would consume time even if we
// weren't single stepping.
static byte
OpZ80_brkpt(word Addr)
{
    return RdZ80_norm(Addr);	// FIXME
}
#endif


static byte
RdZ80_brkpt(word Addr)
{
    byte data = RdZ80_norm(Addr);

    if (BRKPT_ADDR(Addr)) {

	// potentially a hit -- do slow check
	int hit = breakpoint_test(BRK_RD, Addr, (word)data, &Z80Regs.brknum);

	if (hit) {
	    Z80Regs.brkpt = 1;
	    Z80Regs.reason = BRKPT_BRK;
	    Z80Regs.XCount = Z80Regs.ICount;
	    Z80Regs.ICount = 0;	// terminates execution loop (if any)
	}
    }

    return data;
}


static void
WrZ80_brkpt(word Addr, byte Value)
{
    WrZ80_norm(Addr, Value);

    if (BRKPT_ADDR(Addr)) {

	// potentially a hit -- do slow check
	int hit = breakpoint_test(BRK_WR, Addr, (word)Value, &Z80Regs.brknum);

	if (hit) {
	    Z80Regs.brkpt = 1;
	    Z80Regs.reason = BRKPT_BRK;
	    Z80Regs.XCount = Z80Regs.ICount;
	    Z80Regs.ICount = 0;	// terminates execution loop (if any)
	}
    }
}


static byte
InZ80_brkpt(word Port)
{
    byte data = InZ80_norm(Port);

    if (BRKPT_IO(Port)) {

	// potentially a hit -- do slow check
	int hit = breakpoint_test(BRK_IN, Port, (word)data, &Z80Regs.brknum);

	if (hit) {
	    Z80Regs.brkpt = 1;
	    Z80Regs.reason = BRKPT_BRK;
	    Z80Regs.XCount = Z80Regs.ICount;
	    Z80Regs.ICount = 0;	// terminates execution loop (if any)
	}
    }

    return data;
}


static void
OutZ80_brkpt(word Port, byte Value)
{
    OutZ80_norm(Port, Value);

    if (BRKPT_IO(Port)) {

	// potentially a hit -- do slow check
	int hit = breakpoint_test(BRK_OUT, Port, (word)Value, &Z80Regs.brknum);

	if (hit) {
	    Z80Regs.brkpt = 1;
	    Z80Regs.reason = BRKPT_BRK;
	    Z80Regs.XCount = Z80Regs.ICount;
	    Z80Regs.ICount = 0;	// terminates execution loop (if any)
	}
    }
}


#if 0
// this is called by the RunZ80() routine periodically to
// allow system events to be handled.  we return one of
// these values to indicate what the Z80 emulator should do:
//    INT_IRQ   --  Standard RST 38h interrupt
//    INT_NMI   --  Non-maskable interrupt
//    INT_NONE  --  No interrupt required
//    INT_QUIT  --  Exit the emulation

word
LoopZ80(Z80 *R)
{
    return INT_NONE;
}


// special trap -- not a real Z80 feature
void
PatchZ80(Z80 *R)
{
    return;
}
#endif


// ---------------------------------------
//   simulation control
// ---------------------------------------


// execute one instruction and return
void
simstep(void)
{
    int t1;

    Z80Regs.startPC = Z80Regs.PC;

    t1 = Z80Regs.ICount;
    (void)(simstate.exec8080_func)(&Z80Regs);

    TimerTick(t1 - Z80Regs.ICount);
}


// run for Z80Regs.ICount simulated 8080 clock cycles and return
void
simrun(void)
{
    int t1;

    while (Z80Regs.ICount > 0) {
	Z80Regs.startPC = Z80Regs.PC;
	t1 = Z80Regs.ICount;
	(void)(simstate.exec8080_func)(&Z80Regs);
	TimerTick(t1 - Z80Regs.ICount);
    }
}


// step over simstate.stepcount instructions or until ICount
// simulated 8080 cycles have elapsed.
void
simstepover(void)
{
    int t1;
    byte nextop;

    while (Z80Regs.ICount > 0) {

	Z80Regs.startPC = Z80Regs.PC;
	t1 = Z80Regs.ICount;

	if (simstate.runstate == STEP) {
	    nextop = DiZ80(Z80Regs.PC.W);
	    if (  (nextop         == 0xCD) || // call
		 ((nextop & 0xC7) == 0xC4)) { // call conditional
		breakpoint_temp_add((word)(Z80Regs.PC.W + 3));  // just past call
		simstate.runstate = STEPRUN;	// run until breakpoint
	    }
	}

	(void)(simstate.exec8080_func)(&Z80Regs);
	TimerTick(t1 - Z80Regs.ICount);

	if (Z80Regs.brkpt) {

	    if (Z80Regs.brknum == 0) {
		// ok, we're not really done, we've just hit the
		// step-over temporary breakpoint.  note we didn't
		// perform the operation, so don't dec stepcount.
		breakpoint_temp_kill();
		Z80Regs.ICount = Z80Regs.XCount; // restore count
		simstate.runstate = STEP;
		 simstate.stepcount--;
	    } else {
		// we hit a real breakpoint
		return;
	    }

	} else if (simstate.runstate == STEP)
	     simstate.stepcount--;

	if (simstate.stepcount < 1) {
	    // we've stepped the appropriate number of times;
	    // go back to debugger mode
	    Z80Regs.ICount = 0;
	    breakpoint_temp_kill();
	    activate_debugwin_step();
	    simstate.runstate = HALTED;
	    report_curstate();
	    return;
	}

    } // while

    // we ran out of cycles in this time slice, but leave active
    // the STEP state and the remaining number of stepcounts.
}


// report why we've stopped
static void
brkpt_notify(void)
{
    char buf[80];

    // if it was a temporary breakpoint (run-to address, single step),
    // don't squawk about it
    if (Z80Regs.brknum == 0)
	return;

    activate_debugwin();

    switch (Z80Regs.reason) {
	case BRKPT_BRK:
	    sprintf(buf, "  Processor hit breakpoint %d at PC=$%04X.", Z80Regs.brknum, Z80Regs.startPC.W);
	    CmdLog(buf, FALSE);
	    sprintf(buf, "blist %d", Z80Regs.brknum);
	    dbg_interp(buf);
	    break;
	case BRKPT_ILLEGAL:
	    sprintf(buf, "  Processor executed undefined instruction at PC=$%04X.\nReset the Sol.", Z80Regs.PC.W);
	    CmdLog(buf, FALSE);
	    break;
	case BRKPT_HALT:
	    sprintf(buf, "  Processor executed HALT instruction at PC=$%04X.\nReset the Sol.", Z80Regs.PC.W);
	    CmdLog(buf, FALSE);
	    break;
	default:
	    break;
    }
}


// this is called every time we transition to the halted state
static void
report_curstate(void)
{
    char buf[120];
    int len;

    len = sprintf(buf, "  A=%02X F=%c%c%c%c%c%c BC=%04X DE=%04X HL=%04X SP=%04X PC=%04X ",
		      Z80Regs.AF.B.h,
		    ((Z80Regs.AF.B.l) & 0x80) ? 'S' : '-',
		    ((Z80Regs.AF.B.l) & 0x40) ? 'Z' : '-',
		    ((Z80Regs.AF.B.l) & 0x10) ? 'H' : '-',
		    ((Z80Regs.AF.B.l) & 0x04) ? 'P' : '-',
		    ((Z80Regs.AF.B.l) & 0x02) ? 'N' : '-',
		    ((Z80Regs.AF.B.l) & 0x01) ? 'C' : '-',
		      Z80Regs.BC.W,
		      Z80Regs.DE.W,
		      Z80Regs.HL.W,
		      Z80Regs.SP.W,
		      Z80Regs.PC.W
		);

#if 1
    // add disassembly to tail end of buf
    (void)DAsm(&buf[len], (word)Z80Regs.PC.W, 0);
#endif

    CmdLog(buf, FALSE);
}


// this continually runs a timeslice's worth of 8080 cycles, then it
// resynchronizes to the real-world timeslice duration.  it is during
// during this resynchronization that windows messages get handled.
// MAX_CATCHUP defines how many time slices we will attempt to make
// up for, but if we ever fall further behind than that, we just
// live with the fact that we blew it.  This prevents us from running
// virtually unregulated for a long time after some serious speed hit
// (eg, lots of scrolling, or the user does a menu selection).

#define MAX_CATCHUP 4

void
DoTimeSlicing(void)
{
    int         sol_timeslip = 0;	// accumulated time difference
    uint32      x86_timeslip = 0;	// accumulated time difference
    uint32	slice_start, slice_end;
    uint32	sol_ticks_tot, x86_ticks_beg, x86_ticks_tot;
    int		passes;
    int         done;

    simstate.tick_freq = QueryTimebase();	// get system timer frequency

    simstate.timeslice_ms = 10;			// in milliseconds

    if (simstate.tick_freq == 0) {
	WinAlert("This machine doesn't support a high-precision timer.\n" \
		 "Running without speed regulation.\n");
	simstate.up_speed = MHZ_unregulated;
    }
    UpdateTimerInfo();


    slice_start = CurTick();
    passes = 0;
    done = 0;

    startrun();

    while (!done) {

	uint32 tick_diff;
	int i;

	if (Z80Regs.brkpt) {
	    Z80Regs.brkpt = 0;		// don't trip on it again
	    brkpt_notify();
	    breakpoint_temp_kill();	// in case it came from temp breakpoint
	    activate_debugwin();
	    simstate.runstate = HALTED;
	    report_curstate();
	}

	if ((simstate.runstate == RESET) ||
	    (simstate.runstate == HALTED)) {
	    while (!done && MessageLoopGuts(&done))
		;
	    if (simstate.runstate == RESET) {
		if (!WinResetState())
		    simstate.runstate = RUNNING;
	    }
	    continue;
	}


	if (simstate.runstate == STEPI) {
	    int i;
	    simstate.firststep = 1;
	    for(i=0; i<simstate.stepcount; i++) {
		simstep();
		if (Z80Regs.brkpt)
		    break;
	    }
	    activate_debugwin_step();
	    simstate.runstate = HALTED;
	    report_curstate();
	    continue;
	}

	// simulate for one timeslice's worth of 8080 instructions
	if (simstate.tick_freq != 0) {

	    // kick off cassette timers if needed.  it is possible
	    // that both tape players are ON and playing, but this
	    // really isn't expected.  we simply look to see who's
	    // motor is on.
	    if (simstate.vt_rdtimer < 0) {
		int unit = (sysstate.vt[1].motor_on) &&
			   (sysstate.vt[1].playstate != EMPTY) &&
			   (sysstate.vt[1].playstate != RECORD) &&
			   (sysstate.vt[1].playstate != STOP);
		handle_svt_rdtimer(unit, 0);
	    }

	    // kick of tape TX channel if it isn't going
	    if (simstate.vt_wrtimer < 0)
		handle_svt_wrtimer(0, 0);	// kick off

	    // # of 8080 clocks to perform, accounting for a few extra
	    // ticks we may have performed in the previous slice
	    Z80Regs.ICount = simstate.sol_ticks_per_slice + sol_timeslip;

	    if (passes == 0) {
		sol_ticks_tot = 0;
		x86_ticks_tot = 0;
		x86_ticks_beg = CurTick();
	    }
	    sol_ticks_tot += Z80Regs.ICount;
	    passes++;


	    // run for R->ICount 8080 cycles
	    if (simstate.runstate == RUNNING)
		simrun();
	    else
		simstepover();
	    sol_timeslip = Z80Regs.ICount;	// may be negative


	    // detect if PC is in personality module space or not.
	    // this info is used heuristically to know if we should
	    // allow the user to load a program.
	    if ((Z80Regs.PC.W < 0xC000) || (Z80Regs.PC.W > 0xC7FF))
		simstate.prog_detect = 30;	// reset timer
	    else if (simstate.prog_detect > 0)
		simstate.prog_detect--;


	    // drain any messages
	    while (!done && MessageLoopGuts(&done))
		;


	    slice_end = CurTick();
	    tick_diff = ElapsedTick(slice_start,slice_end);
	    x86_ticks_tot += tick_diff;


	    // every so often update the title bar
	    // print out some emulation speed stats
	    if (passes == 200) {
		char buf[80];
		sprintf(buf, "Simulated speed: %6.2f MHz",
			(sol_ticks_tot/1000000.0f*simstate.tick_freq)
		      / (slice_end - x86_ticks_beg));
		if (simstate.up_speed != MHZ_unregulated)
		    sprintf(buf+strlen(buf),"      x86 load: %3d%%",
			(100*x86_ticks_tot)/(passes*simstate.x86_ticks_per_slice));
		StatusBarMsg(1, buf);
		passes = 0;
	    }


	    // wait until the next time slice.
	    // one approach taken was to Sleep() until near the end of the
	    // timeslice, then busywait until the very end.  Sleep() under
	    // win95 was a fiasco.  for instance, Sleep() with 1 ms returned
	    // after anywhere from 0.6ms to 12ms.
	    if (simstate.up_speed != MHZ_unregulated) {
		do {
		    // don't just hog the CPU -- let windows process messages
		    MessageLoopGuts(&done);
		    if (done)
			break;
		    slice_end = CurTick();
		    tick_diff = ElapsedTick(slice_start,slice_end);
		} while(tick_diff < simstate.x86_ticks_per_slice - x86_timeslip);
	    } else {
		slice_end = CurTick();
		tick_diff = ElapsedTick(slice_start,slice_end);
	    }

	    slice_start = slice_end;
	    x86_timeslip = tick_diff - simstate.timeslice_ms;
	    if (x86_timeslip > MAX_CATCHUP*simstate.timeslice_ms) {
		x86_timeslip = MAX_CATCHUP*simstate.timeslice_ms;
	    }

	} else {

	    // we don't have an accurate timer to work with, so just run
	    // unregulated -- simulate 1000 instructions ASAP
	    for(i=0; i<1000 && (!Z80Regs.brkpt); i++)
		simstep();

	    // now drain any messages
	    while (!done && MessageLoopGuts(&done))
		;
	}

	// we could have done this before we started running,
	// but often times (esp single step, or running with
	// breakpoints) we don't get through even one time slice
	// before we halt again, causing excessive flashing.
	// instead, we wait a little bit before disabling the
	// debug window.
	if (simstate.inactivate_timer > 0) {
	    simstate.inactivate_timer--;
	    if (simstate.inactivate_timer == 0)
		inactivate_debugwin();
	}

    } // while(!done)
}


// compute timeslice info after something has changed
void
UpdateTimerInfo(void)
{
    // the Sol timebase is based on a 14.31818 MHz colorburst crystal
    const int colorburst = 14318180;

    switch (simstate.up_speed) {
	case MHZ_2_04:
	    simstate.sol_ticks_per_slice = (int)((simstate.timeslice_ms*colorburst/7) / 1000);
	    simstate.sol_ticks_per_ms    = (int)((colorburst/7) / 1000);
	    TimerTickScale(7);
	    break;
	case MHZ_2_38:
	    simstate.sol_ticks_per_slice = (int)((simstate.timeslice_ms*colorburst/6) / 1000);
	    simstate.sol_ticks_per_ms    = (int)((colorburst/6) / 1000);
	    TimerTickScale(6);
	    break;
	case MHZ_2_86:
	default:
	    simstate.sol_ticks_per_slice = (int)((simstate.timeslice_ms*colorburst/5) / 1000);
	    simstate.sol_ticks_per_ms    = (int)((colorburst/5) / 1000);
	    TimerTickScale(5);
	    break;
    }

    simstate.x86_ticks_per_slice = (simstate.tick_freq*simstate.timeslice_ms)/1000;

#if 1
    printf("timeslice = %d ms\n", simstate.timeslice_ms);
    printf("sol_ticks_per_slice = %6d\n", simstate.sol_ticks_per_slice);
    printf("x86_ticks_per_slice = %6d\n", simstate.x86_ticks_per_slice);
#endif
}


