// Solace -- Sol Anachronistic Computer Emulation
// A Win32 emulator for the Sol-20 computer.
//
// Copyright (c) Jim Battle, 2000, 2001

// various type definitions used by Sol emulator

#ifndef _SOLACE_H_
#define _SOLACE_H_


#if 1
#include <assert.h>
#define ASSERT(x) assert(x)
#else
#define ASSERT(x)
#endif

#include <windows.h>	// only used for DbgWinProc()

#include "vtape.h"
#include "z80.h"


// ============== types ================

typedef unsigned char uint8;
typedef unsigned int  uint32;


typedef enum {
	SW1_RESET	= 0x01,
	SW1_BLANKCTL	= 0x04,
	SW1_POLARITY	= 0x08,
	SW1_BLINKCUR	= 0x10,
	SW1_SOLIDCUR	= 0x20,
} sw1_t;

// uP speed
typedef enum { MHZ_2_04, MHZ_2_38, MHZ_2_86, MHZ_unregulated } speed_t;

// for use by Z80 emulator
typedef byte rdz80fp(word);
typedef void wrz80fp(word, byte);
typedef word exec8080fp(Z80 *);


// simstate holds things related to running the emulator that aren't
// part of the physical Sol system.  bookkeeping type things live here.
typedef struct {

    enum {
	RESET,		// CPU is in reset
	HALTED,		// emulator is not stepping 8080
	RUNNING,	// 8080 is free running until breakpoint
	STEP,		// step over N 8080 instructions, then -> HALTED
	STEPRUN,	// run until temp breakpoint, then -> STEP
	STEPI,		// step N 8080 instructions, then -> HALTED
    } runstate;
    int stepcount;
    int firststep;
    int inactivate_timer;	// how many slices before disabling dbg window

    exec8080fp *exec8080_func;	// normal or debugging routine

    rdz80fp *diz80_func;	// byte RdZ80(word Addr)
    rdz80fp *opz80_func;	// byte RdZ80(word Addr)
    rdz80fp *rdz80_func;	// byte RdZ80(word Addr)
    wrz80fp *wrz80_func;	// void WrZ80(word Addr, byte Value)
    rdz80fp *inz80_func;	// byte InZ80(word Port)
    wrz80fp *outz80_func;	// void OutZ80(word Port, byte Value)

    uint32 timeslice_ms;	// how often we resync
    uint32 tick_freq;		// # timer ticks per second
    speed_t up_speed;		// processor speed
    uint32 sol_ticks_per_ms;	// # of 8080 clocks per millisecond
    uint32 sol_ticks_per_slice;	// # of 8080 clocks per timeslice
    uint32 x86_ticks_per_slice;	// # of x86 clocks per timeslice

    int prog_detect;		// # timeslices since PC was outside SOLOS

    int hscrolltimer;		// timer handle for port 0xFE
    int hscanadv;		// timer handle for port 0xFE
    int hkb_stat;		// timer handle for keyboard strobe timer

    // virtual tape state
    int vt_realtime;		// boolean: 0=ultrafast, 1=real time
    int vt_wrtimer;		// write timer handle
    int vt_rdtimer;		// read timer handle

} simstate_t;


// contains a map per byte of whether there is a special
extern uint32 brkpt_map[2048];	// a bit indicating if anything special is going on
extern byte brkpt_iomap[256];	// is anything special is going on

// types of breakpoints
typedef enum { BRK_INVALID, BRK_TEMP, BRK_OP, BRK_IN, BRK_OUT,
	       BRK_RD, BRK_RD16, BRK_WR, BRK_WR16 } break_t;

// there are a number of breakpoints we support
typedef struct {
    break_t btype;	// type of breakpoint
    int     enabled;	// is breakpoint enabled
    int     sernum;	// breakpoint serial #
    int     addr;	// address to break on
    int     data;	// 8 or 16b data value to trigger on
    int     mask;	// 8 or 16b data mask to trigger on
    int     triggered;	// this breakpoint was detected
    int     dlo;	// rd16/wr16 low  datum (-1 means not touched)
    int     dhi;	// rd16/wr16 high datum (-1 means not touched)
} breakpoint_t;

#define NUM_BREAKPOINTS 32	// # of possible breakpoints
extern breakpoint_t breakpoints[NUM_BREAKPOINTS];
extern int breaknum;		// # of defined breakpoints
extern int breaknxt;		// breakpoint serial number

// the windows interface notifies the core simulator of certain events
#define DEBUG_KEY      1
#define RUN_KEY        2
#define FORCERUN_KEY   3
#define STEP_KEY       4
#define STEPI_KEY      5
#define CLOSE_WIN      6


// display state -- used by winmain.c and windbg.c
// FIXME:windbg.c only looks at hInst and hDbgWnd.
//       perhaps we can move this out of solace.h.
typedef struct {

    HINSTANCE hInst;	// application instance

    // because the DC is private, we don't need to keep getting the
    // window DC, doing a makecompatibleDC, selecting the bitmap,
    // then deleting the DC for every character.  it is way too slow.
    // instead, we do it once then just use this copy of it.
    HDC  hdcWin;	// = GetDC(winstate.hWnd);

    HWND hWnd;		// handle to sol screen
    int  exitcode;	// what WM_QUIT returned (ugly hack)

    HWND hDbgWnd;	// handle to debugger window

    HWND hOpts;		// handle to options propsheet

    DWORD cellwidth;	// # pixels per font character
    DWORD cellheight;	// # pixels per font character

    DWORD snapcnt;	// snapshot count

    BOOL useHPTimer;		// high-precision timer is available
    BOOL useLPTimer;		// high-precision timer is available
    DWORD tick_freq;		// # timer ticks per second

    // pick a monitor color
    enum { DISP_WHITE, DISP_GREEN, DISP_AMBER } disp_color;
    COLORREF fg, bg;		// corresponding fg and bg colors

    int hidemouse;	// hide mouse if it is in client area
    int stats;		// show stats on titlebar
    int holdmsg;	// # time slices to not update titlebar

    int timerflag;	// gets set when it is time to change blink state
    int blinkmask;	// cursor blink gate

    HBITMAP hCharmap;	// handle to font bitmap
    HDC  hdcCharMap;	// = CreateCompatibleDC(hdcWin);

    // used for searching for files
    //   [0] is used for searching out ROMs (.rom)
    //   [1] is used for searching out programs (.ent)
    OPENFILENAME ofn[2];
    char basedir[_MAX_PATH];	// directory we launched from

    enum { mainonly, mainnow, altnow } fontmode; // choose font
    byte *pFont;	// pointer to character generator ROM
    byte *pMainFont;	// pointer to character generator ROM
    byte *pAltFont;	// pointer to character generator ROM
    int mapwidth;	// font table width,  in pixels
    int mapheight;	// font table height, in pixels
    int mapstride;	// font table width,  in pixels

    HCURSOR hNullCursor;
    HCURSOR hArrowCursor;

    int scrollshade;	// scroll blanking
    int scrollbase;	// scroll offset

    // ------ the following params are screen geometry information ------

    int client_w;	// current window width
    int client_h;	// current window height

    int scrn_w;		// # pixels in display
    int scrn_h;		// # pixels in display
    int scrn_stride;	// # bytes per row of image

    int xoff;		// screen overscan, in pixels
    int yoff;		// screen overscan, in pixels

    HANDLE hStatus;	// handle to status bar
    int status_h;	// status bar height

} winstate_t;
extern winstate_t winstate;

// character generators
extern uint8 charset_6574[];
extern uint8 charset_6575[];
extern uint8 charset_dzonkha[];
extern uint8 charset_devanagri[];

// built-in ROM
extern byte solos_rom[];

// unprincipled debugging global
int g_foo;


// ============== exported by solace.c ================

// really should move this to solace.c and make it private,
// but for now, winmain uses scrollbase stuff frequently for blitting.

// playstate
enum { EMPTY, STOP, PLAY, FWD, REW, RECORD };

typedef struct {

    // OUT 0FEH state:
    int scrolltimer;	// out port 0xFE tweaked within last 250 ms
    int scrollshade;	// how many beginning rows to blank
    int scrollbase;	// which line # is the first displayed
    int scanadv;	// horizontal sync (more or less)

    // this semaphore and register represent the keyboard status and
    // input data latch.  it is filled by winmain and drained by emulator.
    int kb_stat;	// 0=no key, 1=key
    int kb_key;

    sw1_t dipsw1;	// display dip swith
    byte  dipsw2;	// sense switches
    byte  dipsw3;	// serial port baud rate
    byte  dipsw4;	// serial port control bits

    // virtual tape state
    int   vt_TTBE;		// tape TX buffer empty
    int   vt_writebyte;		// if !TTBE, this is data to write
    int   vt_rxready;		// boolean: 1=we already have sample
    int   vt_readbyte;		// if rxready, what is data
    int   vt_framing;		// if rxready, 1=framing error
    int   vt_overrun;		// if rxready, 1=overrun error
    int   vt_baud;		// 300 or 1200
    struct {
	int playstate;		// current tape mode
	int motor_on;		// boolean: 1=tape motor is enabled
	vtape_t *tape;		// tape units 0 and 1
    } vt[2];

} sysstate_t;


// this holds state that has a physical correlate in the Sol system
extern sysstate_t sysstate;

// this holds state that is emulator bookkeeping information
extern simstate_t simstate;	// exported because debugger.c needs it

// this is the memory image.  parts of it might
// be ghosted by other functions.
extern byte memimage[65536];

// initialize system state
void InitEmu(void);	  // state unaffected by reset; called once
void SysReset(int warm);  // state affected by reset
void LoadDefaultROM(void);

// register keyboard activity
void SysKeyPress(byte data);

// certain windows events
void notify_coresim(int event, int param);
void bkptchange_coresim(void);

// run one timeslice
void DoTimeSlicing(void);

// set and get various properties
void SetDipSwitch(int swtch, byte value);
byte GetDipSwitch(int swtch);
void SetClockSpeed(int kilohertz);
int  GetClockSpeed(void);
void    Set8080Speed(speed_t clkrate);
speed_t Get8080Speed(void);
int ProgramRunning(void);

// get/set state of indicated unit
enum {	TPROP_POSITION	   = VTPROP_POSITION,
	TPROP_IGNOREBAUD   = VTPROP_IGNOREBAUD,
	TPROP_WRITEPROTECT = VTPROP_WRITEPROTECT,
	TPROP_DIRTY	   = VTPROP_DIRTY,
	TPROP_BAUDRATE	   = VTPROP_BAUDRATE,
	TPROP_LABEL   	   = VTPROP_LABEL,
	TPROP_REALTIME=7,
	TPROP_PLAYSTATE=8 };
void GetTapeProp(int unit, int prop, int *param);
void SetTapeProp(int unit, int prop, int param);

enum {VTF_NewTape, VTF_LoadTape, VTF_SaveTape, VTF_EjectTape};
int TapeFile(int unit, int action, char *filename);


// ============== exported by winmain.c ================

// draw one specific character on the display
void blit_char_to_disp(int row, int col, byte data);

// redraw a subset of characters on the screen
void update_disp(int all, int inv, int ctl);

// send a message to the status bar for some of (approximate) seconds
void WinNote(char *msg, int duration);
// send a message to the status bar
void StatusBarMsg(int subwin, char *msg);
// send an error/warning to the user
void WinAlert(char *msg);

// reset simulation
void WinResetSim(int warm_reset);

// return a realtime counter of unknown resolution
uint32 QueryTimebase(void);
uint32 CurTick(void);
uint32 ElapsedTick(uint32 begin, uint32 end);

// process one windows message, if one is pending
int MessageLoopGuts(int *done);

// change the color of the emulated display tube
void SetDispColors(void);

// change the display character generator
void CreateCharMap(void);
int WinOutFFTickle(void);

// update the screen after the scrollshade register changes
void WinUpdateScrollBase(int scrollshade, int scrollbase);

// filename functions
char *extract_path(char *name);
int absolute_path(char *name);


// ============== exported by winmain_kb.c ================

void OnKeyDown(HWND hwnd, UINT vk, BOOL down, int repeat, UINT flags);
void OnSetFocus(HWND hwnd, HWND oldwin);
void OnKillFocus(HWND hwnd, HWND newwin);
int WinResetState(void);
void WinInitKB(void);
int My_Handle_WM_CHAR(WPARAM wParam, LPARAM lParam);


// ============== exported by winmain_opts.c ================

void InitOpts(void);
void CreateOptsPropSheet(void);


// ============== exported by windbg.c ================

void InitDebugger(void);
BOOL CALLBACK DbgWinProc(HWND hDlg, UINT iMsg, WPARAM wParam, LPARAM lParam);
void activate_debugwin_step(void);
void inactivate_debugwin(void);
void activate_debugwin(void);
void destroy_debugwin(void);
int active_debugwin(void);
void CmdLog(char *str, int bold);

void update_debugwin(int update_flags);
#define UPDATE_REG	0x0001	// registers subwindow
#define UPDATE_MEM	0x0002	// memory subwindow
#define UPDATE_DASM	0x0004	// dasm subwindow
#define UPDATE_LOG	0x0008	// command log subwindow
#define UPDATE_CMD	0x0010	// command entry subwindow
#define HOME_DASM	0x0100	// recenter dasm around pc
#define HOME_MEM	0x0200	// recenter mem around sp

#define UPDATE_REFRESH (UPDATE_REG | UPDATE_MEM | UPDATE_DASM | UPDATE_LOG | UPDATE_CMD)
#define UPDATE_ALL     (UPDATE_REFRESH | HOME_DASM | HOME_MEM)


// ============== exported by Z80.c ================

// z80 state
extern Z80 Z80Regs;


// ============== exported by scheduler.c ================

typedef void (*timercb_t)(uint32 arg1, uint32 arg2);
void TimerInit(void);
int TimerCreate(uint32 ticks, timercb_t fcn, uint32 arg1, uint32 arg2);
void TimerKill(int n);
void TimerTickScale(int n);
void TimerTick(uint32 n);


// ============== exported by Debug.c ================

int DAsm(char *S, word A, int extended);
int OpLen(word A);


// ============== exported by debugger.c ================

void breakpoint_init(void);
void breakpoint_genmap(void);
void breakpoint_temp_add(word addr);
void breakpoint_temp_kill(void);
int  breakpoint_test(break_t btype, word addr, word data, int *num);
int  breakpoint_verify(int *num);
int any_breakpoints(break_t btype);
int check_brk(word addr, break_t type);
void dbg_interp(char *str);


// ============== exported by fileio.c ================

void read_bin_file(char *filename, int report);
int create_hexdump(char *filename, int start, int end, byte *memory);
int create_ihexdump(char *filename, int start, int end, byte *memory);


#endif // ifdef _SOLACE_H_
