// Solace -- Sol Anachronistic Computer Emulation
// A Win32 emulator for the Sol-20 computer.
//
// Copyright (c) Jim Battle, 2000, 2001

// some ideas for enhancement:
//  *) add trace [nn] and tracei [nn] commands.  like step/stepi,
//     except it logs instruction and register state into log window
//     before/after? each step/stepi.
//     or perhaps a "trace on/off" command that causes logging as
//     each instruction is executed in a subsequent step/stepi.
//  *) keep track of register/memory state at each breakpoint
//     and use this info to draw changed values in red
//  *) add ability to log output of window:
//        (NEW)LOG <filename>, LOG OFF, LOG ON
//  *) allow resizing of panes (at least between dasm and log windows)
//  *) allow reading in a symbol table;
//     use for expressions and disassembly
//  *) allow specifying a source .prn and use this for
//     disassembly when appropriate

// The registers window has a few different states:
//     1) non-existant
//     2) active   (when the CPU is stopped)
//     3) inactive (showing, but only the CLOSE and STOP buttons work)
//     4) hidden   (not shown)
//
// we are notified by the core emulator what to do via the

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <windowsx.h>	// message cracker macros

#include "z80.h"
#include "solace.h"
#include "resource.h"
#include "windbg_sub.h"


// overall debugger state
dbgstate_t dbgstate;

static void hide_debugwin(void);	// forward declaration


// ===============================================================
// this routine is used by any subwindow to gray out the
// area if the window is disabled
// ===============================================================

void
GrayOutWindow(HWND hwnd, HDC hdc)
{
    // color the display gray & disable scrollbars
    COLORREF oldcolor;
    HBRUSH newbrush, oldbrush;
    HPEN oldpen, newpen;
    RECT rc;

    COLORREF bg = RGB(0xD0, 0xD0, 0xD0);

    GetClientRect(hwnd, &rc);

    oldcolor = SetBkColor(hdc, bg);
    newbrush = CreateSolidBrush(bg);
    newpen   = CreatePen(PS_SOLID, 1, bg);
    oldbrush = SelectObject(hdc, newbrush);
    oldpen   = SelectObject(hdc, newpen);

    Rectangle(hdc, 0, 0, rc.right, rc.bottom);	// l, t, r, b

    SetBkColor(hdc, oldcolor);
    SelectObject(hdc, oldbrush);
    SelectObject(hdc, oldpen);
    DeleteObject(newbrush);
    DeleteObject(newpen);
}


void
handle_fkey(UINT key)
{
    switch (key) {
	case VK_F5: notify_coresim(RUN_KEY, 0); break;
	case VK_F6: notify_coresim(STEP_KEY, 1); break;
	case VK_F7: notify_coresim(STEPI_KEY, 1); break;
	default: break;
    }
}


// ==================================================================
// the following are related to the main debugger window
// ==================================================================

HWND
CreateDbgWindow(void)
{
    // typical windows window
    DWORD winstyle = WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN;

    HWND hwnd;

    // open up debugging container window
    hwnd = CreateWindow(
		"SolaceDbgWindow",	// window class name
		"Sol Debug",		// window caption
		winstyle,		// window style
		CW_USEDEFAULT,		// initial x position
		0,			// initial y position
		CW_USEDEFAULT,		// initial x size
		0,			// initial y size
		NULL,			// parent window handle
		NULL,			// window menu handle
		winstate.hInst,		// program instance handle
		NULL);			// creation parameters

    ASSERT(hwnd != NULL);

    return hwnd;
}


static BOOL
OnCreate(HWND hwnd, LPCREATESTRUCT lpCreateStruct)
{
    HDC hdc = GetDC(hwnd);
    int saved = SaveDC(hdc);

    int PointSize = 8;
    int nHeight = -MulDiv(PointSize, GetDeviceCaps(hdc, LOGPIXELSY), 72);

    dbgstate.cour_font = CreateFont(
			nHeight,	// height
			0,		// width (default)
			0,0,		// escapement, orientation
			FW_REGULAR,	// weight
			0,0,0,		// italic,underline,strikeout
			ANSI_CHARSET,	// maybe DEFAULT_CHARSET
			OUT_RASTER_PRECIS,
			CLIP_DEFAULT_PRECIS,
			DEFAULT_QUALITY,
			FIXED_PITCH | FF_MODERN,
			"Courier New"
			);

    dbgstate.fixed_font = (dbgstate.cour_font == NULL) ? GetStockFont(ANSI_FIXED_FONT)
						       : dbgstate.cour_font;

    // get some basic information about the font we are using
    {
	TEXTMETRIC fix_tm;		// info about the fixed font

	if (dbgstate.fixed_font != NULL)
	    SelectFont(hdc, dbgstate.fixed_font);

	GetTextMetrics(hdc, &fix_tm);

	dbgstate.fix_x = fix_tm.tmAveCharWidth;
	dbgstate.fix_y = fix_tm.tmHeight + fix_tm.tmExternalLeading;
    }


    // if it was opened & closed previously, use the size/location
    // from when it was last closed.
    if (dbgstate.placement_valid) {
	SetWindowPlacement(hwnd, &dbgstate.placement);
    } else {
	// guess a "good" initial size; leave position, but change size.
	// reg window is 16 chars wide, plus two border pixels;
	// the cmd window should be about 72 wide, plus three borders
	// of four pixels apiece, plus some pad.
	int width  = (16+72+2)*dbgstate.fix_x + 3*4 + 10;
	int height = 480;
	SetWindowPos(hwnd, HWND_TOP, 0, 0, width, height, SWP_NOMOVE);
    }

    RestoreDC(hdc, saved);
    ReleaseDC(hwnd, hdc);

    // create subordinate windows
    dbgstate.hRegWnd  = CreateRegWindow(hwnd);
    dbgstate.hMemWnd  = CreateMemWindow(hwnd);
    dbgstate.hDasmWnd = CreateDasmWindow(hwnd);
    dbgstate.hCmdWnd  = CreateCmdWindow(hwnd);
    dbgstate.hLogWnd  = CreateLogWindow(hwnd);

    update_debugwin(UPDATE_ALL);

    return TRUE;
}


// handle window resize events
// we mainly just compute what the sizes of the subordinate
// windows should be and send them messages to resize.
//
// the layout is as follows:
//   +-----+-------------------------------------+
//   |     |                                     |
//   |     |  dasm window                        |
//   |regs |                                     |
//   |     |-------------------------------------+
//   |     |                                     |
//   |-----|  command output (hLogWnd)           |
//   |     |                                     |
//   | mem |                                     |
//   |     |-------------------------------------|
//   |     |  command prompt (hCmdWnd)           |
//   +-----+-------------------------------------+
//
// the size of regs and command prompt are fixed;
// dasm and command output windows split what is left,
// and stack is as tall as it can be.

static void
OnSize(HWND hwnd, UINT state, int cx, int cy)
{
    int reg_top, reg_left, reg_width, reg_height;
    int mem_top, mem_left, mem_width, mem_height;
    int cmd_top, cmd_left, cmd_width, cmd_height;
    int dasm_left, dasm_width, dasm_height, dasm_top;
    int log_left, log_width, log_height, log_top;
    int border = 2;

    // the two left windows are of fixed width:
    // width is 16 characters wide:
    //   DE=hhhh=ddddd
    //   STACK nnnn:hhhh

    reg_top    = border;
    reg_left   = 0;
    reg_width  = dbgstate.fix_x * 16;	// 16 chars wide
    reg_height = dbgstate.fix_y *  8;	//  8 chars tall

    mem_top    = reg_top + reg_height + border;
    mem_left   = 0;
    mem_width  = reg_width;
    mem_height = cy - mem_top;

    cmd_left   = reg_left + reg_width + border;
    cmd_width  = cx - cmd_left;
    cmd_height = (dbgstate.fix_y * 3)/2;
    cmd_top    = cy - cmd_height;

    dasm_top    = border;
    dasm_left   = cmd_left;
    dasm_width  = cmd_width;
    dasm_height = (cmd_top - dasm_top - 2*border)/2;

    log_left   = dasm_left;
    log_top    = dasm_top + dasm_height + border;
    log_width  = dasm_width;
    log_height = cmd_top - border - (dasm_top + dasm_height + border);
    
    MoveWindow(dbgstate.hRegWnd,  reg_left,  reg_top,  reg_width,  reg_height,  TRUE);
    MoveWindow(dbgstate.hMemWnd,  mem_left,  mem_top,  mem_width,  mem_height,  TRUE);
    MoveWindow(dbgstate.hDasmWnd, dasm_left, dasm_top, dasm_width, dasm_height, TRUE);
    MoveWindow(dbgstate.hCmdWnd,  cmd_left,  cmd_top,  cmd_width,  cmd_height,  TRUE);
    MoveWindow(dbgstate.hLogWnd,  log_left,  log_top,  log_width,  log_height,  TRUE);
}


static void
OnPaint(HWND hwnd)
{
    PAINTSTRUCT	ps;
    HDC hdc = BeginPaint(hwnd, &ps);
    int saved = SaveDC(hdc);
    
    RestoreDC(hdc, saved);
    EndPaint(hwnd, &ps);
}


// the user has requested that the debug window be closed
static void
OnClose(HWND hwnd)
{
#if 1
    // just hide the window, don't destroy it
    hide_debugwin();
    notify_coresim(FORCERUN_KEY, 0);
#else
    // if we do this, we will cause the command and log windows
    // to be destroyed, losing their state.  next time the debugger
    // is invoked, they'd have no history.
    notify_coresim(CLOSE_WIN, 0);
#endif
}


// destroy the debugger window
static void
OnDestroy(HWND hwnd)
{
    // get the position of the debug window so that if it is
    // opened again, we can open it in the same location
    dbgstate.placement.length = sizeof(WINDOWPLACEMENT);
    GetWindowPlacement(hwnd, &dbgstate.placement);
    dbgstate.placement_valid = TRUE;

    if (dbgstate.cour_font != NULL) {
	DeleteObject(dbgstate.cour_font);
	dbgstate.cour_font = NULL;
    }
}


// debug window handler
static LRESULT CALLBACK
WndProcDbg(HWND hWnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
    switch (iMsg) {
	HANDLE_MSG(hWnd, WM_CREATE,  OnCreate);
	HANDLE_MSG(hWnd, WM_SIZE,    OnSize);
	HANDLE_MSG(hWnd, WM_PAINT,   OnPaint);
	HANDLE_MSG(hWnd, WM_CLOSE,   OnClose);
	HANDLE_MSG(hWnd, WM_DESTROY, OnDestroy);

	case WM_SETFOCUS:
	    SetFocus(dbgstate.hCmdWnd);
	    break;
    }

    return DefWindowProc(hWnd, iMsg, wParam, lParam);
}


static void
RegisterDbgClass(void)
{
    WNDCLASSEX wndclass;

    wndclass.cbSize        = sizeof(wndclass);
    wndclass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
    wndclass.lpfnWndProc   = WndProcDbg;
    wndclass.cbClsExtra    = 0;
    wndclass.cbWndExtra    = 0;
    wndclass.hInstance     = winstate.hInst;
    wndclass.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
    wndclass.hCursor       = NULL; //LoadCursor(NULL, IDC_ARROW);
    wndclass.hbrBackground = (HBRUSH)(COLOR_3DFACE + 1);
    wndclass.lpszMenuName  = NULL;
    wndclass.lpszClassName = "SolaceDbgWindow";
    wndclass.hIconSm       = LoadIcon(NULL, IDI_APPLICATION);

    RegisterClassEx(&wndclass);
}


// indicate if debugger window is active or not
int
active_debugwin(void)
{
    return dbgstate.active;
}


// if the debug window doesn't exist, create it.  if it exists,
// make sure all controls are reactivated.
void
activate_debugwin(void)
{
    if (winstate.hDbgWnd == NULL) {
	winstate.hDbgWnd = CreateDbgWindow();		// create debug window
    }
    ShowWindow(winstate.hDbgWnd, SW_SHOW);

    dbgstate.active = TRUE;

    // enable subwindows
    InvalidateRect(dbgstate.hRegWnd, NULL, TRUE);
    DasmWindowAction(WA_ENABLE, 0);
    MemWindowAction(WA_ENABLE, 0);
    LogWindowAction(WA_ENABLE, 0);

    // place current SP in middle of memory window
    MemWindowAction(WA_GOTO, Z80Regs.SP.W);

    update_debugwin(UPDATE_ALL);

    SetFocus(dbgstate.hCmdWnd);	// cmd edit window gets focus
}


// inactivate all the subwindows
void
inactivate_debugwin(void)
{
    if ((winstate.hDbgWnd == NULL) || !dbgstate.active)
	return;

    dbgstate.active = FALSE;

    // disable subwindows
    InvalidateRect(dbgstate.hRegWnd, NULL, TRUE);
    DasmWindowAction(WA_DISABLE, 0);
    MemWindowAction(WA_DISABLE, 0);
    // LogWindowAction(WA_DISABLE, 0);  nah, leave it going

    // since program is running again, move focus to main window
    SetFocus(winstate.hWnd);
}


// hide the debugger window, keeping its state active
static void
hide_debugwin(void)
{
    if (winstate.hDbgWnd == NULL)
	return;

    ShowWindow(winstate.hDbgWnd, SW_HIDE);

    dbgstate.active = FALSE;
}


// if the debug window doesn't exist, create it.  if it exists,
// make sure all controls are reactivated.
void
destroy_debugwin(void)
{
    if (winstate.hDbgWnd != NULL) {
	DestroyWindow(winstate.hDbgWnd);
	winstate.hDbgWnd = NULL;
    }

    dbgstate.active = FALSE;
}


void
update_debugwin(int update_flags)
{
    if (winstate.hDbgWnd == NULL)
	return;

    if (update_flags & HOME_DASM)
	DasmWindowAction(WA_GOTO, Z80Regs.PC.W);
    if (update_flags & HOME_MEM)
	MemWindowAction(WA_GOTO,  Z80Regs.SP.W);

    if (update_flags & UPDATE_REG)
	InvalidateRect(dbgstate.hRegWnd,  NULL, TRUE);
    if (update_flags & UPDATE_DASM)
	InvalidateRect(dbgstate.hDasmWnd, NULL, TRUE);
    if (update_flags & UPDATE_MEM)
	InvalidateRect(dbgstate.hMemWnd,  NULL, TRUE);
    if (update_flags & UPDATE_CMD)
	InvalidateRect(dbgstate.hCmdWnd,  NULL, TRUE);
    if (update_flags & UPDATE_LOG)
	InvalidateRect(dbgstate.hLogWnd,  NULL, TRUE);

    UpdateWindow(winstate.hDbgWnd);		// cause redraw of everything
}


// like update_debugwin(), but used in single-stepping situations
// to prevent redrawing all windows (less flashing)
void
activate_debugwin_step(void)
{
    if (winstate.hDbgWnd == NULL)
	return;

    MemWindowAction(WA_GOTO,  Z80Regs.SP.W);
    DasmWindowAction(WA_GOTO, Z80Regs.PC.W);

    InvalidateRect(dbgstate.hRegWnd,  NULL, TRUE);
    InvalidateRect(dbgstate.hDasmWnd, NULL, TRUE);
    InvalidateRect(dbgstate.hMemWnd,  NULL, TRUE);
    UpdateWindow(winstate.hDbgWnd);		// cause redraw of everything

    SetFocus(dbgstate.hCmdWnd);	// cmd edit window gets focus
}


// ==================================================================
// this is called once at program start-up time
// ==================================================================

void
InitDebugger(void)
{
    dbgstate.active          = FALSE;
    dbgstate.placement_valid = FALSE;
    dbgstate.cour_font       = NULL;
    dbgstate.hDasmWnd        = 0;
    dbgstate.hRegWnd         = 0;

    InitEditHist();

    DasmWindowAction(WA_INIT, 0);
    MemWindowAction(WA_INIT, 0);
    LogWindowAction(WA_INIT, 0);

    RegisterDbgClass();		// register main debugger window

    RegisterRegClass();		// register main debugger window
    RegisterDasmClass();	// register disassembler window
    RegisterMemClass();		// register memory display window
    RegisterCmdClass();		// register command input window
    RegisterLogClass();		// register command output window
}

