// Solace -- Sol Anachronistic Computer Emulation
// A Win32 emulator for the Sol-20 computer.
//
// Copyright (c) Jim Battle, 2000, 2001

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <windows.h>
#include <windowsx.h>	// message cracker macros
#include <io.h>		// required for open_osfhandle() function
#include <fcntl.h>	// needed for stdio redirection hack
#include <direct.h>	// needed for _getcwd() etc
#include <mmsystem.h>	// needed for timeGetTime() and related funcs
#include <commctrl.h>	// needed for status bar

#include "solace.h"	// one glob of everything
#include "wintape.h"	// interface to virtual tape drives
#include "hexfile.h"	// ENT and HEX file I/O
#include "resource.h"

#define BLINK_RATE 3.0	// in Hz (actually, blink period is half of this)


// ------ globals ------

// state relevant to the windows environment
winstate_t winstate;

// .ini file name (sans directory)
const char *g_ini_file = "\\solace.ini";


// -------- forward references --------

static void RegisterWindowClass(HINSTANCE hInstance, char *appname);
static void CreateMainWindow(HINSTANCE hInstance, char *appname);
static LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);
static LRESULT CALLBACK CreditsDlgProc(HWND, UINT, WPARAM, LPARAM);
static LRESULT CALLBACK KeysDlgProc(HWND, UINT, WPARAM, LPARAM);
static void BuildStatusBar(HWND hwnd);
static void UpdateMenu(void);
static void InitWinstate(void);
static void InitOpenDlg(void);
static BOOL FileOpenDlg(HWND hWnd, int romfile);
static void CreateConsole(void);
static void CreateSnapshot(void);
static void LoadOptions(void);
static void SaveOptions(void);
static char *MainFontName(void);
static char *AltFontName(void);


int WINAPI
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
	PSTR szCmdLine, int iCmdShow)
{
    static char szAppName[] = "Solace";

    if (0)
	CreateConsole();	// create console window for printf messages

    InitWinstate();		// winstate.foo = blah
    InitOpts();			// options dialog

    RegisterWindowClass(hInstance, szAppName);	// register main window class
    CreateMainWindow(hInstance, szAppName);	// create main window

    InitDebugger();		// one-time initializations
    InitEmu();			// set some emulator state
    InitOpenDlg();		// set up "Open" dialog params
    WinTapeInit(hInstance, winstate.hWnd); // stuff related to virtual tape
    LoadOptions();		// load .ini file to override defaults
    WinResetSim(FALSE);		// cold reset the simulator
    SetDispColors();		// set foreground/background colors for display
    CreateCharMap();		// create font map
    notify_coresim(FORCERUN_KEY, 0);

    ShowWindow(winstate.hWnd, iCmdShow);	// not visible until this point
    UpdateWindow(winstate.hWnd);		// cause redraw of everything

    DoTimeSlicing();

    return winstate.exitcode;
}


static void
RegisterWindowClass(HINSTANCE hInstance, char *appname)
{
    WNDCLASSEX  wndclass;

    wndclass.cbSize        = sizeof(wndclass);
    wndclass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
    wndclass.lpfnWndProc   = WndProc;
    wndclass.cbClsExtra    = 0;
    wndclass.cbWndExtra    = 0;
    wndclass.hInstance     = hInstance;
    //wndclass.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
    wndclass.hIcon         = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON1));
    wndclass.hCursor       = NULL; //LoadCursor(NULL, IDC_ARROW);
    //wndclass.hbrBackground = (HBRUSH) GetStockObject(BLACK_BRUSH);
    wndclass.hbrBackground = NULL;
    wndclass.lpszMenuName  = MAKEINTRESOURCE(IDR_MENU1);
    wndclass.lpszClassName = appname;
    wndclass.hIconSm       = NULL;

    RegisterClassEx(&wndclass);
    winstate.hInst = hInstance;
}


static void
CreateMainWindow(HINSTANCE hInstance, char *appname)
{
    DWORD	winstyle;

    // open up main window
#if 1
    // typical windows window
    winstyle = WS_OVERLAPPEDWINDOW;
#else
    // don't allow resizing of the window
    winstyle = WS_OVERLAPPED | WS_CAPTION;
#endif

    winstate.hWnd = CreateWindow(
			appname,	// window class name
			"Solace",	// window caption
			winstyle,	// window style
			CW_USEDEFAULT,	// initial x position
			0,		// initial y position
			CW_USEDEFAULT,	// initial x size
			0,		// initial y size
			NULL,		// parent window handle
			NULL,		// window menu handle
			hInstance,	// program instance handle
			NULL);		// creation parameters

    ASSERT(winstate.hWnd != NULL);

    winstate.hdcWin = GetDC(winstate.hWnd);
}


static void
InitWinstate(void)
{
    winstate.hWnd         = NULL;	// WinResetSim looks at it before it is set
    winstate.hdcWin       = NULL;	// = GetDC(winstate.hWnd);
    winstate.hdcCharMap   = NULL;	// = CreateCompatibleDC(hdcWin);
    winstate.hCharmap     = NULL;

    winstate.hOpts        = NULL;

    winstate.exitcode     = 0;
    winstate.hNullCursor  = 0;
    winstate.hArrowCursor = 0;

    winstate.fontmode     = mainonly;
    winstate.pMainFont    = charset_6574;
    winstate.pAltFont     = charset_devanagri;
    winstate.pFont        = winstate.pMainFont;
                          
    winstate.cellwidth    = 9;
    winstate.cellheight   = 13*2;
    winstate.mapwidth     = 32;	// the cell is narrower
    winstate.mapheight    = 256*winstate.cellheight;
                          
    winstate.scrn_w       = 64*winstate.cellwidth;
    winstate.scrn_h       = 16*winstate.cellheight;
    winstate.scrn_stride  = winstate.scrn_w/8;
                          
    winstate.disp_color   = DISP_WHITE;
    winstate.hidemouse    = 0;
    winstate.stats        = 0;
    winstate.holdmsg      = 0;
                          
    winstate.timerflag    = 0;
    winstate.blinkmask    = 1;	// 1 = on
                          
    winstate.scrollbase   = 0;
    winstate.scrollshade  = 0;
                          
    winstate.snapcnt      = 0;
                          
    winstate.client_w     = 0;	// determined on WM_SIZE
    winstate.client_h     = 0;
                          
    winstate.hStatus      = NULL;
    winstate.status_h     = 0;	// determined on WM_CREATE
}

#define StatusID (200)		// arbitrary (?) unique constant


// this is called any time the processor updates this register
// from chapter 8 of the sol manual:
//
//    U2 latches the starting row address from DIO0-3 and U13
//    latches the data on DIO4-7, with !PORT_OUT_FE being the
//    strobe.  Data on DIO4-7 specifies where the first line will
//    be displayed.  Thus, the number loaded into U1 is the
//    address of the first displayable scan line, and the number
//    loaded into U22 defines the character row (0 through 15).
//
// in other words, bits [7:4] specify that the first 'n' otherwise
// displayable lines are blank, and [3:0] specify the row number of
// the first actually displayed line.
//
// Note: SOLOS uses only [3:0], and bits [7:4] are always 0.
void
WinUpdateScrollBase(int scrollshade, int scrollbase)
{
    if ((scrollbase  == winstate.scrollbase) &&
        (scrollshade == winstate.scrollshade))
	return;

    // changing scrollbase causes a roll.  Ideally, we would
    // blit the small part to a temp buffer, then do a screen-
    // to-screen blit of the large part, then do a mem-to-screen
    // blit of the small part we saved.
    //
    // we don't, though.  for now, just handle the most typical case.
    if ( (1==1) &&
         (scrollshade == winstate.scrollshade) &&
         (scrollshade == 0) &&
         (winstate.yoff > 0) &&	// all screen rows are visible
	(((scrollbase-winstate.scrollbase) & 0xF) == 1)) {
	RECT rc;
	int r,c;

	// in case screen is clipped on the bottom
	int bot = min(winstate.yoff + winstate.scrn_h,
		      winstate.client_h - winstate.status_h);

	rc.left   = winstate.xoff;
	rc.right  = winstate.xoff + winstate.scrn_w;
	rc.top    = winstate.yoff;
	rc.bottom = winstate.yoff + bot;

	// scroll up scroll one row
	ScrollWindowEx(winstate.hWnd,
			0, -(int)winstate.cellheight,
			&rc, NULL, NULL, NULL,
			0);	// don't scroll statusbar

	// draw newly exposed area (assuming all 16 rows are visible)
	// the proper way to do this is to pass SW_INVALIDATE as the
	// final param to ScrollWindowEX(), which causes the newly
	// exposed area to be invalidating, resulting in a call to
	// OnPaint with a cliprect equal to the exposed area.  However,
	// since the OnPaint handler is stupid and just draws everything,
	// it would kill the speed advantage we are trying to reap.
	// Another lesson in why dirty code begets dirty code...
	winstate.scrollbase  = scrollbase;
	winstate.scrollshade = scrollshade;

	r = (scrollbase-1) & 0xF;
	for(c=0; c < 64; c++) {
	    byte data = memimage[0xCC00 + r*64 + c];
	    blit_char_to_disp(r, c, data);
	}

    } else {
	// just redraw the whole window
	winstate.scrollbase  = scrollbase;
	winstate.scrollshade = scrollshade;
	update_disp(1,0,0);
    }

}


// if the cursor toggle flag has been seen since the last
// time slice, redraw the appropriate parts of the screen.
// also, the status message area gets blanked after a while.
static void
UpdateCursorBlinkState(void)
{
    if (winstate.timerflag) {

	winstate.timerflag = 0;

	winstate.blinkmask = !winstate.blinkmask;
	if (GetDipSwitch(1) & SW1_BLINKCUR) {
	    update_disp(0,1,0);
	}

	// after displaying the status message for some
	// period, erase it
	if (winstate.holdmsg > 0) {
	    winstate.holdmsg--;
	    if (winstate.holdmsg == 0)
		StatusBarMsg(0, "");
	}
    }
}


// if any messages are waiting, dispatch one and return TRUE.
// the *done flag is is set to 1 if we see a WM_QUIT message.
int
MessageLoopGuts(int *done)
{
    MSG msg;

    // update cursor blink state
    UpdateCursorBlinkState();

    if (!PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
	return FALSE;

    if (winstate.hOpts == NULL || !PropSheet_IsDialogMessage(winstate.hOpts, &msg)) {
	TranslateMessage(&msg);
	DispatchMessage(&msg);
    }

    if (msg.message == WM_QUIT) {
	winstate.exitcode = msg.wParam;
	*done = TRUE;
    }

    return TRUE;
}


static BOOL
OnCreate(HWND hwnd, LPCREATESTRUCT lpCreateStruct)
{
    int menuheight = GetSystemMetrics(SM_CYMENU);	// in pixels
    int cx_border  = GetSystemMetrics(SM_CXSIZEFRAME);	// in pixels
    int cy_border  = GetSystemMetrics(SM_CYSIZEFRAME);	// in pixels
    int window_w, window_h;
    RECT rcwin;

    // we use two different mouse cursors at different times
    winstate.hNullCursor  = LoadCursor(NULL, IDC_ICON);
    winstate.hArrowCursor = LoadCursor(NULL, IDC_ARROW);

    BuildStatusBar(hwnd);	// self evident

    // now that we know the statusbar height, compute the size of
    // the entire client area (the screen plus the status bar).

    window_w = winstate.scrn_w		// the Sol display canvas
	     + 6			// add a little border around screen
	     + 2*cx_border;		// left and right window borders

    window_h = menuheight		// FIXME: titlebar (fudged)
	     + menuheight		// height of menu bar
	     + winstate.scrn_h		// the Sol display canvas
	     + 6 			// add a little border around screen
	     + winstate.status_h	// height of status bar
	     + 2*cy_border;		// top and bottom window borders

    // preserve location, but change size
    GetWindowRect(hwnd, &rcwin);
    (void)MoveWindow(hwnd, rcwin.left, rcwin.top, window_w, window_h, FALSE);

    // get messages provide for cursor blinking
    SetTimer(hwnd, 1, (int)(1000/BLINK_RATE), NULL);

    return TRUE;
}


// handle window resize events
static void
OnSize(HWND hwnd, UINT state, int cx, int cy)
{
    winstate.client_w = cx;	// width of client area
    winstate.client_h = cy;	// height of client area

    // if the window is bigger than the display,
    // center the display in the window
    winstate.xoff = (cx - winstate.scrn_w)/2;
    winstate.yoff = (cy - winstate.status_h - winstate.scrn_h)/2;
    if (winstate.xoff < 0) winstate.xoff = 0;
    if (winstate.yoff < 0) winstate.yoff = 0;

    // wndclass has CS_HREDRAW | CS_VREDRAW, so we don't need to:
    //    InvalidateRect(winstate.hWnd, NULL, TRUE);

    // let the status bar know the size has changed
    SendMessage(winstate.hStatus, WM_SIZE, (WPARAM)0, (LPARAM)0);
}


static void
OnPaint(HWND hwnd)
{
    PAINTSTRUCT	ps;
    HDC         hdc    = BeginPaint(hwnd, &ps);
    HBRUSH      hBrush = CreateSolidBrush(winstate.bg);
    HPEN        hPen   = CreatePen(PS_SOLID, 1, winstate.bg);
    HBRUSH      hOldBrush;
    HPEN        hOldPen;

    // clear the overscan around the display
    // set the brush to the background color for filling
    hOldBrush = SelectObject(hdc, hBrush);
    hOldPen   = SelectObject(hdc, hPen);

    // top strip
    (void)Rectangle( hdc,
	  0,			// left
	  0,			// top
	  winstate.client_w,	// right
	  winstate.yoff		// bottom
    );
    // bottom strip
    (void)Rectangle( hdc,
	  0,					// left
	  winstate.yoff + winstate.scrn_h,	// top
	  winstate.client_w,			// right
	  winstate.client_h			// bottom
    );
    // left strip
    (void)Rectangle( hdc,
	  0,					// left
	  winstate.yoff,			// top
	  winstate.xoff,			// right
	  winstate.yoff + winstate.scrn_h	// bottom
    );
    // right strip
    (void)Rectangle( hdc,
	  winstate.xoff + winstate.scrn_w,	// left
	  winstate.yoff,			// top
	  winstate.client_w,			// right
	  winstate.yoff + winstate.scrn_h	// bottom
    );

    // update all the chars of the display
    if (simstate.runstate == RESET)
	(void)Rectangle( hdc,
	      winstate.xoff,			// left
	      winstate.yoff,			// top
	      winstate.client_w,		// right
	      winstate.yoff + winstate.scrn_h);	// bottom
    else
	update_disp(1,0,0);

    // clean up
    SelectObject(hdc, hOldBrush);
    SelectObject(hdc, hOldPen);
    DeleteObject(hBrush);
    DeleteObject(hPen);
    EndPaint(hwnd, &ps);
}


static void
OnCommand(HWND hwnd, int id, HWND hctl, UINT codenotify)
{
    switch (id) {

    // --- File submenu ---

    case IDM_FILE_TAPE1:
	CreateTapeWindow(1);
	break;

    case IDM_FILE_TAPE2:
	CreateTapeWindow(2);
	break;

    case IDM_FILE_LOADROM:
	if (FileOpenDlg(hwnd, TRUE)) {
	    // always reset after changing the ROM
	    WinResetSim(FALSE);
	}
	return;

    case IDM_FILE_LOADPROGRAM:
	if (ProgramRunning()) {
	    int confirm = MessageBox(hwnd,
		"A program is already running.  Are you sure\n" \
		"you want to load another? (maybe reset first)",
		"Confirm", MB_OKCANCEL | MB_ICONEXCLAMATION);
	    if (confirm != IDOK)
		return;
	}
	if (FileOpenDlg(hwnd, FALSE)) {
	    (void)hexfile_read(winstate.ofn[1].lpstrFile, 1, memimage);
	    // just in case the program loaded into screen memory,
	    // refresh the display (one program cutely has its
	    // ascii instructions loaded directly to the screen)
	    update_disp(1, 0, 0);
	}
	return;

    case IDM_FILE_SNAPSHOT:
	CreateSnapshot();
	return;

    case IDM_FILE_SAVEOPTIONS:
	SaveOptions();
	return;

    case IDM_FILE_EXIT:
	SendMessage(hwnd, WM_CLOSE, 0, 0L);
	return;


    // --- Processor submenu ---

    case IDM_PROC_COLDBOOT:
	WinResetSim(FALSE);	// cold reset
	return;

    case IDM_PROC_WARMBOOT:
	WinResetSim(TRUE);	// warm reset
	return;

    case IDM_PROC_2_04MHZ:
	Set8080Speed(MHZ_2_04);
	return;

    case IDM_PROC_2_38MHZ:
	Set8080Speed(MHZ_2_38);
	return;

    case IDM_PROC_2_86MHZ:
	Set8080Speed(MHZ_2_86);
	return;

    case IDM_PROC_999MHZ:
	Set8080Speed(MHZ_unregulated);
	return;


    // --- Display submenu ---

    case IDM_DISP_WHITE:
	winstate.disp_color = DISP_WHITE;
	SetDispColors();
	return;

    case IDM_DISP_GREEN:
	winstate.disp_color = DISP_GREEN;
	SetDispColors();
	return;

    case IDM_DISP_AMBER:
	winstate.disp_color = DISP_AMBER;
	SetDispColors();
	return;

    case IDM_DISP_HIDEMOUSE:
	winstate.hidemouse = !winstate.hidemouse;
	// don't need to do anything else.  as soon as the user moves the
	// mouse, the cursor will disappear on the WM_MOUSEMOVE msg.
	return;

    case IDM_DISP_STATS:
	if (winstate.stats)
	    StatusBarMsg(1, "");
	winstate.stats = !winstate.stats;
	return;


    // --- Options... submenu ---
    case IDM_OPTIONS:
	if (winstate.hOpts == NULL)
	    CreateOptsPropSheet();
	SetFocus(winstate.hOpts);
	break;


    // --- Debug... submenu ---
    case IDM_DEBUG:
	notify_coresim(RUN_KEY, 0);
	break;


    // --- HELP submenu ---

    case IDM_HELP_CREDITS:
	(void)DialogBox(winstate.hInst, MAKEINTRESOURCE(IDD_HELP_CREDITS),
			hwnd, (DLGPROC)CreditsDlgProc);
	return;

    case IDM_HELP_KEYS:
	(void)DialogBox(winstate.hInst, MAKEINTRESOURCE(IDD_HELP_KEYS),
			hwnd, (DLGPROC)KeysDlgProc);
	return;

    case IDM_HELP_DEBUG:
	{
	    char buf[_MAX_PATH];
	    strcpy(buf, winstate.basedir);
	    strcat(buf, "\\html");
	    (void)ShellExecute(NULL, NULL, "solace_dbg.html", NULL, buf, 0);
	}
	return;

    default:
	break;
    }

    // we aren't handling it, pass it through to default handler
    FORWARD_WM_COMMAND(hwnd, id, hctl, codenotify, DefWindowProc);
}


// potentially hide mouse cursor when it is over the Sol screen
static void
OnMouseMove(HWND hwnd, int xClient, int yClient, UINT keyflags)
{
    if (winstate.hidemouse)
	SetCursor(winstate.hNullCursor);
    else
	SetCursor(winstate.hArrowCursor);
}


// update the checkmarks, etc, of the popup menus
static void
OnInitMenuPopup(HWND hwnd, HMENU hmenu, UINT item, BOOL sysmenu)
{
    if (!sysmenu)
	UpdateMenu();
}


static void
OnMenuSelect(HWND hwnd, HMENU hmenu, int item, HMENU hpopup, UINT flags)
{
    if ((hmenu == NULL) && (flags == 0xFFFFFFFF)) {
	// the menu has been dismissed
	StatusBarMsg(0, "");
	return;
    }

    if (!(flags & (MF_SYSMENU | MF_POPUP))) {
	// offer help only for non-system, non-popup menus;
	// that is, our own submenu items
	char buf[80];
	if (item == 40003)
	    sprintf(buf, "Save an image of the screen in snap%04d.bmp", winstate.snapcnt);
	else
	    (void)LoadString(winstate.hInst, item, buf, sizeof(buf));
	StatusBarMsg(0, buf);
    }
}


// blink timer interrupt
static void
OnTimer(HWND hwnd, UINT id)
{
    winstate.timerflag = 1;
}


// a request has been made to shut down.  we might honor it, or not.
static void
OnClose(HWND hwnd)
{
    if (WinTapeClose())
	return;		// cancelled

    // OK, go ahead and shut down
    DestroyWindow(winstate.hWnd);
}


// terminate the application, including killing children.
// WM_DESTROY kills proper child winows (in this case, the Options window)
// but because the tape windows and the debugger window aren't children,
// we must explicitly clobber them.
static void
OnDestroy(HWND hwnd)
{
    WinTapeDestroy();	// includes killing kids

    if (winstate.hDbgWnd != NULL)
	DestroyWindow(winstate.hDbgWnd);

    KillTimer(hwnd, 1);
    PostQuitMessage(0);
}


// main window handler
static LRESULT CALLBACK
WndProc(HWND hWnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
    switch (iMsg) {

	HANDLE_MSG(hWnd, WM_CREATE,        OnCreate);
	HANDLE_MSG(hWnd, WM_SIZE,          OnSize);
	HANDLE_MSG(hWnd, WM_PAINT,         OnPaint);
	HANDLE_MSG(hWnd, WM_MOUSEMOVE,     OnMouseMove);
	HANDLE_MSG(hWnd, WM_KEYDOWN,       OnKeyDown);
	HANDLE_MSG(hWnd, WM_INITMENUPOPUP, OnInitMenuPopup);
	HANDLE_MSG(hWnd, WM_MENUSELECT,    OnMenuSelect);
	HANDLE_MSG(hWnd, WM_COMMAND,       OnCommand);
	HANDLE_MSG(hWnd, WM_TIMER,         OnTimer);
	HANDLE_MSG(hWnd, WM_CLOSE,         OnClose);
	HANDLE_MSG(hWnd, WM_DESTROY,       OnDestroy);
	HANDLE_MSG(hWnd, WM_SETFOCUS,      OnSetFocus);
	HANDLE_MSG(hWnd, WM_KILLFOCUS,     OnKillFocus);

	// we can't use a standard cracker because it doesn't give us
	// the transition state we need to defeat auto repeat
	case WM_CHAR:
	    return My_Handle_WM_CHAR(wParam, lParam);
    }

    return DefWindowProc(hWnd, iMsg, wParam, lParam);
}


static LRESULT CALLBACK
CreditsDlgProc(HWND hDlg, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
    char *webpage = "http://www.thebattles.net/sol20/sol.html";

    switch (iMsg) {

	case WM_INITDIALOG:
	    return TRUE;

	case WM_COMMAND:
	    switch (LOWORD(wParam)) {
		case IDOK:
		    EndDialog(hDlg, LOWORD(wParam));
		    return TRUE;
		case IDCANCEL:
		    EndDialog(hDlg, LOWORD(wParam));
		    return TRUE;
		case IDC_BUTTON_SOLPAGE:
		    (void)ShellExecute(NULL, NULL, webpage, NULL, NULL, SW_SHOWNORMAL);
		    return TRUE;
		default:
		    break;
	    }
	    break;

	default:
	    break;

    } // switch (iMsg)

    return FALSE;	// message not handled
}


static LRESULT CALLBACK
KeysDlgProc(HWND hDlg, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
    switch (iMsg) {

	case WM_INITDIALOG:
	    return TRUE;

	case WM_COMMAND:
	    switch (LOWORD(wParam)) {
		case IDOK:
		    EndDialog(hDlg, LOWORD(wParam));
		    return TRUE;
		case IDCANCEL:
		    EndDialog(hDlg, LOWORD(wParam));
		    return TRUE;
		default:
		    break;
	    }
	    break;

	default:
	    break;

    } // switch (iMsg)

    return FALSE;	// message not handled
}


// we maintain state for search out ROMs and for searching out programs.
// it is possible that they exist in two different places.
static void
InitOpenDlg(void)
{
    char *fullpath[2], *initdir[2];
    int i;

    (void)_getcwd(winstate.basedir, _MAX_PATH);

    for(i=0; i<2; i++) {

	fullpath[i]  = (char*)malloc(_MAX_PATH + 1);
	initdir[i]   = (char*)malloc(_MAX_PATH + 1);
	ASSERT(fullpath[i] != NULL);
	ASSERT(initdir[i]  != NULL);

	memset(&winstate.ofn[i], 0x00, sizeof(OPENFILENAME));

	winstate.ofn[i].lStructSize = sizeof(OPENFILENAME);
	winstate.ofn[i].lpstrFilter = (i) ? "ENT Files (*.ent)\0*.ENT\0" \
					    "HEX Files (*.hex)\0*.HEX\0" \
					    "All Files (*.*)\0*.*\0\0"
					  : "ROM Files (*.rom)\0*.ROM\0" \
					    "HEX Files (*.hex)\0*.HEX\0" \
					    "All Files (*.*)\0*.*\0\0";
	winstate.ofn[i].nFilterIndex    = 1;
	winstate.ofn[i].lpstrFile       = fullpath[i];
	winstate.ofn[i].nMaxFile        = _MAX_PATH;
	winstate.ofn[i].lpstrInitialDir = initdir[i];
	winstate.ofn[i].Flags           = OFN_HIDEREADONLY | OFN_FILEMUSTEXIST;
	winstate.ofn[i].lpstrDefExt     = (i) ? "ent" : "rom";

	// set the default search directory to where we launched from
	strcpy(initdir[i], winstate.basedir);
	strcat(initdir[i], (i) ? "\\binaries" : "\\ROMs");

	// default filename to open
	fullpath[i][0] = '\0';
    }

    // get ROM from rom directory
    strcpy(fullpath[0], initdir[0]);
    strcat(fullpath[0],"\\solos.rom");	// default PM ROM

  for(i=0; i<2; i++) {
     printf("init: ofn[%d].lpstrFile       = %s\n", i, winstate.ofn[i].lpstrFile);
     printf("init: ofn[%d].lpstrInitialDir = %s\n", i, winstate.ofn[i].lpstrInitialDir);
  }
}


static BOOL
FileOpenDlg(HWND hWnd, int romfile)
{
    BOOL flag;
    int idx = !romfile;
    char *prevfile = strdup(winstate.ofn[idx].lpstrFile);

    winstate.ofn[idx].hwndOwner = hWnd;
    winstate.ofn[idx].lpstrFile[0] = '\0';

    // although the presense of lpstrInitialDir is supposed to specify
    // which dir to open a file in, it doesn't always work for me.
    // so, I explicitly change to that dir before opening the file.
#if 0
    if (strlen(winstate.ofn[idx].lpstrInitialDir) > 0) {
	flag = _chdir(winstate.ofn[idx].lpstrInitialDir);
	if (!flag) {
	    char msg[256];
	    sprintf(msg, "Error: couldn't cd to '%s'\n"
			 "       Using current directory.",
		    winstate.ofn[idx].lpstrInitialDir);
	    WinAlert(msg);
	    strcpy((char*)winstate.ofn[idx].lpstrInitialDir[0], "");
	}
    }
#endif

    flag = GetOpenFileName(&winstate.ofn[idx]);

    // if we succeed in opening a file, we want to remember what
    // directory to start the search in next time.
    if (flag) {
	char *newpath = extract_path(winstate.ofn[idx].lpstrFile);
	strcpy((char*)winstate.ofn[idx].lpstrInitialDir, newpath);
	free(newpath);
    } else {
	// restore it to what it was in case user hit CANCEL button
	if (prevfile != NULL)
	    strcpy(winstate.ofn[idx].lpstrFile, prevfile);
    }

    free(prevfile);
    return flag;
}


// create a console window so printf messages come out
static void
CreateConsole(void)
{
    int hCrt, i;
    FILE *hf;

    AllocConsole();
    hCrt = _open_osfhandle((long)GetStdHandle(STD_OUTPUT_HANDLE), _O_TEXT);
    hf = _fdopen( hCrt, "w" );
    *stdout = *hf;
    i = setvbuf( stdout, NULL, _IONBF, 0 );
}


void
WinResetSim(int warm_reset)
{
    SysReset(warm_reset);		// init Sol state
    if (!warm_reset) {
	int stat = hexfile_read(winstate.ofn[0].lpstrFile, 0, memimage);	// get personality module ROM
	if (stat != HFILE_OK) {
	    WinAlert("The personality ROM image couldn't be loaded.\n"
		     "Using the built-in standard SOLOS ROM.");
	    LoadDefaultROM();
	}
    }
    if (winstate.hWnd)
	InvalidateRect(winstate.hWnd, NULL, TRUE);
}


// the Sol character generator uses a fixed character cell grid of
// 8 bits wide by 11 rows high.  The screen is 64 chars wide by 24
// rows, so the pixel grid on the Sol is 512x264.  We don't want
// to duplicate this exactly because the pixels on our display are
// approximately square, while those on the Sol aren't.  The typical
// TV screen is 4:3.  We simply double each row in our font -- it
// ends up being too tall, but it is better than the squat look.
//
// we create a 1bpp bitmap, 16 chars wide and 8 chars tall.  Sol
// defines only 128 chars, with the top bit specifying inverse chars.

void
CreateCharMap(void)
{
    uint8 *charmap;
    int    mapbytes;
    int    chr, r;

    // release existing objects/handles
    if (winstate.hCharmap != NULL) {
	DeleteObject(winstate.hCharmap);
	winstate.hCharmap = NULL;
    }
    if (winstate.hdcCharMap != NULL) {
	DeleteDC(winstate.hdcCharMap);
	winstate.hdcCharMap = NULL;
    }

    // determine which font is currently active
    switch (winstate.fontmode) {
	case altnow:
	    winstate.pFont = winstate.pAltFont;
	    break;
	case mainnow:
	case mainonly:
	    winstate.pFont = winstate.pMainFont;
	    break;
	default:
	    ASSERT(0);
    }

    // this is the actual size of each character on the display
    winstate.cellwidth  = 9;
    winstate.cellheight = 13*2;

    // this is the size of each character in the font table
    winstate.mapwidth   = 32;	// the cell is narrower
    winstate.mapheight  = 256*winstate.cellheight;
    winstate.mapstride  = winstate.mapwidth/8;

    // allocate space for the expanded bitmap
    mapbytes = winstate.mapheight * winstate.mapstride;
    charmap = (uint8*)malloc(mapbytes);
    if (charmap == NULL) {
	printf("died allocating charmap\n");
	exit(-1);
    }
    memset(charmap,            0xFF, mapbytes/2);	// first 128 are normal
    memset(charmap+mapbytes/2, 0x00, mapbytes/2);	// second 128 are inverse

    // pre-expand the charset into the charmap.
    // also, 0s are foreground, 1's are background, so we must invert.
    // each char begins on a 16 bit boundary.
    for(chr=0; chr<256; chr++) {
	for(r=0; r<13; r++) {
	    byte pix = winstate.pFont[13*(chr&0x7F) + r];
	    charmap[winstate.mapstride*(26*chr + 2*r + 0)] =
	    charmap[winstate.mapstride*(26*chr + 2*r + 1)] =
						(chr < 128) ? ~pix : pix;
	}
    }

    // cache new DC and charmap
    // create windows data structure for the character map
    winstate.hCharmap = CreateBitmap(winstate.mapwidth,
				     winstate.mapheight,
				     1, 1,
				     charmap);
    winstate.hdcCharMap = CreateCompatibleDC(winstate.hdcWin);
    SelectObject(winstate.hdcCharMap, winstate.hCharmap);

    free(charmap);

    update_disp(1,0,0);	// refresh entire display
}


// every time OUT 0FFH is performed, this function
// is called from the core emulator.  it returns
// '1' if we are in DH mode, and 0 otherwise.
int
WinOutFFTickle(void)
{
    switch (winstate.fontmode) {
	case mainnow:
	    winstate.fontmode = altnow;
	    CreateCharMap();
	    return 1;
	case altnow:
	    winstate.fontmode = mainnow;
	    CreateCharMap();
	    return 1;
    }
    return 0;
}


// set screen fg/bg colors
void
SetDispColors(void)
{
    COLORREF fg,bg,tmp;

    if (winstate.hWnd) {

	HDC hdc = GetDC(winstate.hWnd);
	int iBitsPixel = GetDeviceCaps(hdc, BITSPIXEL);

	switch (winstate.disp_color) {
	    case DISP_WHITE: fg = RGB(0xFF,0xFF,0xFF); break;
	    case DISP_GREEN: fg = RGB(0x10,0xF0,0x10); break;
	    case DISP_AMBER: fg = RGB(0xD0,0xD0,0x00); break;
	}
	bg = RGB(0x00,0x00,0x00);
	if (!(GetDipSwitch(1) & SW1_POLARITY)) {
	    tmp = fg; fg = bg; bg = tmp;
	}

	// if we have a palettized display, we must pick
	// the nearest color otherwise it might get dithered
	if (iBitsPixel <= 8) {
	    fg = GetNearestColor(hdc, fg);
	    bg = GetNearestColor(hdc, bg);
	}

	SetTextColor(hdc, fg);
	SetBkColor  (hdc, bg);
	SetBkMode(hdc, OPAQUE); // this is the default anyway

	// cache these
	winstate.fg = fg;	// used by snapshot function
	winstate.bg = bg;	// used by snapshot function

	InvalidateRect(winstate.hWnd, NULL, TRUE);
    }
}


// draw one specific character on the display given the (row,col)
// of the memory image, before display, so scrollbase is removed
// to figure out where it shows up in the display.
void
blit_char_to_disp(int row, int col, byte data)
{
    int inverted = (data & 0x80);
    int dip1 = GetDipSwitch(1);
    int left, upper, height;
    int disprow;

    ASSERT(row >=0 && row <= 15);
    ASSERT(col >=0 && col <= 63);

    // the display is relative to the scroll register
    disprow = (row - sysstate.scrollbase + sysstate.scrollshade) & 0xF;

    left   = winstate.xoff + col*winstate.cellwidth;
    upper  = winstate.yoff + disprow*winstate.cellheight;
    height = winstate.cellheight;

    if (upper + height > winstate.client_h - winstate.status_h) {
	height -= (upper + height) - (winstate.client_h - winstate.status_h);
	if (height < 0)
	    return;
    }

    if (dip1 & SW1_BLANKCTL) {
	// control characters turn into blanks
	if ((data & 0x7F) < 0x20)
	    data = inverted | 0x20;	// space, preserve polarity
    }

    if ( inverted &&
	! (dip1 & SW1_SOLIDCUR) &&
	!((dip1 & SW1_BLINKCUR) && winstate.blinkmask)) {
	    data &= 0x7F;
    }

    // if we are above scrollshade, blank it
    if (disprow < winstate.scrollshade)
	data = 0x20;

    BitBlt(winstate.hdcWin,		// dest DC
	   left, upper,			// upper left corner of dest
	   winstate.cellwidth,		// width
	   height,			// height
	   winstate.hdcCharMap,		// source DC
	   0,				// left ...
	   data*winstate.cellheight,	// ... upper
	   SRCCOPY);			// ROP
}


// updates the entire screen, but only for the characters of the given type:
//    all = redraw all chars
//    inv = redraw only those with bit 7 set
//    ctl = redraw only control characters (inverted or not)
void
update_disp(int all, int inv, int ctl)
{
    byte data;
    int r,c;

    if (all) {
	for(r=0; r < 16; r++) {
	    for(c=0; c < 64; c++) {
		data = memimage[0xCC00 + (r<<6) + c];
		blit_char_to_disp(r, c, data);
	    }
	}
    } else {
	for(r=0; r < 16; r++) {
	    for(c=0; c < 64; c++) {
		data = memimage[0xCC00 + (r<<6) + c];
		if ( (inv && (data >= 0x80)) ||
		     (ctl && ((data & 0x7F) < 0x20))
		   )
		    blit_char_to_disp(r, c, data);
	    }
	}
    }
}


// update checks on menu items
// if we don't have timer control, grey out the speed choices
static void
UpdateMenu(void)
{
    HMENU hMenu = GetMenu(winstate.hWnd);
    HMENU hSubmenu = GetSubMenu(hMenu, 1);	// processor popup menu
    BOOL regulated = winstate.useHPTimer || winstate.useLPTimer;

    EnableMenuItem(hSubmenu, IDM_PROC_2_04MHZ, (regulated) ? MFS_ENABLED : MFS_DISABLED);
    EnableMenuItem(hSubmenu, IDM_PROC_2_38MHZ, (regulated) ? MFS_ENABLED : MFS_DISABLED);
    EnableMenuItem(hSubmenu, IDM_PROC_2_86MHZ, (regulated) ? MFS_ENABLED : MFS_DISABLED);

    CheckMenuItem(hSubmenu, IDM_PROC_2_04MHZ, (Get8080Speed() == MHZ_2_04)        ? MFS_CHECKED : MFS_UNCHECKED);
    CheckMenuItem(hSubmenu, IDM_PROC_2_38MHZ, (Get8080Speed() == MHZ_2_38)        ? MFS_CHECKED : MFS_UNCHECKED);
    CheckMenuItem(hSubmenu, IDM_PROC_2_86MHZ, (Get8080Speed() == MHZ_2_86)        ? MFS_CHECKED : MFS_UNCHECKED);
    CheckMenuItem(hSubmenu, IDM_PROC_999MHZ,  (Get8080Speed() == MHZ_unregulated) ? MFS_CHECKED : MFS_UNCHECKED);

    hSubmenu = GetSubMenu(hMenu, 2);	// display popup menu
    CheckMenuItem(hSubmenu, IDM_DISP_WHITE, (winstate.disp_color == DISP_WHITE) ? MFS_CHECKED : MFS_UNCHECKED);
    CheckMenuItem(hSubmenu, IDM_DISP_GREEN, (winstate.disp_color == DISP_GREEN) ? MFS_CHECKED : MFS_UNCHECKED);
    CheckMenuItem(hSubmenu, IDM_DISP_AMBER, (winstate.disp_color == DISP_AMBER) ? MFS_CHECKED : MFS_UNCHECKED);

    CheckMenuItem(hSubmenu, IDM_DISP_HIDEMOUSE, (winstate.hidemouse) ? MFS_CHECKED : MFS_UNCHECKED);

    EnableMenuItem(hSubmenu, IDM_DISP_STATS, (regulated) ? MFS_ENABLED : MFS_DISABLED);
    CheckMenuItem(hSubmenu,  IDM_DISP_STATS, (winstate.stats) ? MFS_CHECKED : MFS_UNCHECKED);
}


// put up a message on the title bar for some number of time slices
void
WinNote(char *msg, int duration)
{
    StatusBarMsg(0, msg);
    // we use the cursor blink timer to know how long to keep message up
    winstate.holdmsg = (int)(duration*BLINK_RATE+0.8);
}


// if something goes wrong and the only thing to do is report an
// error, this is the way to do it.
void
WinAlert(char *msg)
{
    (void)MessageBox(winstate.hWnd, msg, "Warning", MB_OK | MB_ICONEXCLAMATION);
}


// scan the section of the memimage that corresponds to the display
// and return a bitmap that corresponds to the image it generates.
// a pointer to the bitmap to write to is passed in.  it must be of
// the correct size; if the pointer is NULL, a map is allocated of
// the correct size.  in either case, the pointer to the map is returned.
// note that the sysstate.scrollbase value must be taken into account.
static byte *
CreateScrnImage(byte *imgmap)
{
    // 9b field masks for 8 possible alignments
    //                 off =   0      1      2      3      4      5      6      7
    static byte maskL[8] = { 0xFF, 0x7F, 0x3F, 0x1F, 0x0F, 0x07, 0x03, 0x01 };
    static byte maskR[8] = { 0x80, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC, 0xFE, 0xFF };
    DWORD dstride_1 = (winstate.scrn_stride >> 2);	// one display row
    DWORD dstride_2 = (winstate.scrn_stride >> 1);	// two display rows
    int   blank_controls = GetDipSwitch(1) &  SW1_BLANKCTL;
    int   any_cursor     = GetDipSwitch(1) & (SW1_SOLIDCUR | SW1_BLINKCUR);
    int   row, col, imgbytes;

    imgbytes = winstate.scrn_h * winstate.scrn_stride;

    if (imgmap == NULL) {
	// allocate space for the display bitmap
	imgmap = (byte*)malloc(imgbytes);
	if (imgmap == NULL) {
	    printf("died allocating imgmap in creat_scrnimg\n");
	    exit(-1);
	}
    }


    // initialize the array to the background color
    memset(imgmap, 0xFF, imgbytes);


    for(row=0; row<16; row++) {

	int r = row * 13;	// font table cell height
	int blank_shade = (row < sysstate.scrollshade);
	int effective_row = (row + sysstate.scrollbase - sysstate.scrollshade) & 0xF;
	byte *pData = &memimage[0xCC00 + 64*effective_row];

	for(col=0; col<64; col++, pData++) {

	    int c   = col * 9;	// font table cell width, although only b[6:0] are used
	    int off = (c & 31);	// 32 possible character alignments in DWORD

	    byte data    = *pData;
	    int inverted = (data & 0x80);

	    byte *bp, *cp, *ce, t;
	    DWORD pix;

	    if (inverted && !any_cursor)
		inverted = 0;

	    data &= 0x7F;

	    if ( blank_shade ||
		(blank_controls && (data < 0x20)) )
		// control characters turn into blanks
		data = 0x20;

#if 0
	    // speedup -- however, it makes the time very
	    // display dependent
	    if (data == 0x20 && !inverted)
		continue;
#endif

	    bp = imgmap + r*winstate.scrn_stride*2 + (c>>3);
	    cp = &winstate.pFont[13*data];
	    ce = cp + 13;	// end of char


	    if (off < 24 && (1==1)) {

		// 9b char map doesn't span 32b boundary
		DWORD *dp = (DWORD *)((DWORD)bp & ~3);
		DWORD mmask = ~ (0x1FF << (23-off));
		DWORD nmask = (inverted) ? 0x00000000 : 0xFFFFFFFF;

		off = 24-off;	// later loop invarient

		while (cp<ce) {

		    // get current row of font, align to bit position
		    // possible invert fg/bg, and merge background mask
		    pix = ((*cp << off) ^ nmask) | mmask;

		    // swizzle pixels due to the fact that the format
		    // MS chose for monochrome bitmaps was not the best
		    // for little endian machines.  this is because
		    // bits 7 and 8 of a word are non-contiguous in the
		    // image.  why did they do that?
#if 0
		    pix = ((pix & 0x000000FF) << 24)
		        | ((pix & 0x0000FF00) <<  8)
		        | ((pix & 0x00FF0000) >>  8)
		        | ((pix & 0xFF000000) >> 24);
#else
		    __asm mov eax, pix
		    __asm bswap eax
		    __asm mov pix, eax
#endif
		    // merge pixels into imagemap
		    *(dp + dstride_1) = *dp = (*dp & pix);

		    dp += dstride_2;	// advance two display rows
		    cp++;		// advance one character row

		} // while(cp < ce)

	    } else {

		// the 9b character row straddles a 32b boundary
		// (col * 9) & 7 == ((col * 8) + col) mod 8 == (col & 7)
		off = (col & 7);	// 8 possible character alignments

		while(cp<ce) {

		    // get current row of font
		    pix = *cp << 8;	// leftmost bit is bit 15
		    if (!inverted)
			pix = ~pix;
		    else
			pix = pix | 0xFFFF007F;

		    // left half of 9b
#if 0
		    // this is required if we are updating against a
		    // not-necessarily blank background
		    t = ((*bp) & ~maskL[off]) | ((pix >> (8+off)) & maskL[off]);
#else
		    t = *bp & (byte)(pix >> (8+off));
#endif
		    *(bp + winstate.scrn_stride) = *bp = t;

		    // right half of 9b
		    bp++;
#if 0
		    t = ((*bp) & ~maskR[off]) | ((pix >> off) & maskR[off]);
#else
		    t = *bp & (byte)(pix >> off);
#endif
		    *(bp + winstate.scrn_stride) = *bp = t;

		    bp += 2*winstate.scrn_stride - 1;
		    cp++;

		} // while (cp < ce)
	    }

	} // for col
    } // for row

    return imgmap;
}


static RGBQUAD
COLORREF_TO_RGBQUAD(COLORREF color)
{
    RGBQUAD v;

    v.rgbBlue     = (BYTE)((color >> 16) & 0xFF);
    v.rgbGreen    = (BYTE)((color >>  8) & 0xFF);
    v.rgbRed      = (BYTE)((color >>  0) & 0xFF);
    v.rgbReserved = (BYTE)0x00;

    return v;
}


// copy screen to a file.  on each successive call, a new
// filename is generate by simply incrementing a counter.
static void
CreateSnapshot(void)
{
    BITMAPFILEHEADER bmfh;
    BITMAPINFOHEADER bmih;
    RGBQUAD          aColors[2];

    byte *imgmap, *ip, *ep;
    char  filename[_MAX_PATH];
    int   imgbytes, r;
    FILE *fp;

    // allocate space for the display bitmap
    imgbytes = winstate.scrn_h * winstate.scrn_stride;
    imgmap = (byte*)malloc(imgbytes);
//{
//    DWORD t1, t2, t3, t4;
//    t1 = CurTick();
//    t2 = CurTick();
    (void)CreateScrnImage(imgmap);
//    t3 = CurTick();
//    t4 = (t3-t2) - (t2-t1);
//    printf("time to generate image: %d ticks, or %f uS\n", t4, (t4*1000000.0f)/winstate.tick_freq);
//}
// it consistently took 7.1 ms on my machine, a P55C-200

    // write out a file
    bmfh.bfType      = 'MB';	// "BM" (bitmap)
    bmfh.bfSize      = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER)
		     + 2*sizeof(RGBQUAD) + imgbytes;
    bmfh.bfReserved1 = 0;
    bmfh.bfReserved2 = 0;
    bmfh.bfOffBits   = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER)
		     + 2*sizeof(RGBQUAD);

    bmih.biSize          = sizeof(BITMAPINFOHEADER);
    bmih.biWidth         =  winstate.scrn_w;	// in pixels
    bmih.biHeight        =  winstate.scrn_h;// in pixels, top down
    bmih.biPlanes        = 1;
    bmih.biBitCount      = 1;			// bpp
    bmih.biCompression   = BI_RGB;		// no compression
    bmih.biSizeImage     = 0;			// windows figures it out
    bmih.biXPelsPerMeter = 0;
    bmih.biYPelsPerMeter = 0;
    bmih.biClrUsed       = 0;			// 0 = all
    bmih.biClrImportant  = 0;			// 0 = all

    // the BitBlt command uses 1 as bg and 0 as fg, so we swap
    // the expected mapping to match BitBlt
    aColors[0] = COLORREF_TO_RGBQUAD(winstate.fg);
    aColors[1] = COLORREF_TO_RGBQUAD(winstate.bg);

    sprintf(filename,"%s\\snap%04d.bmp", winstate.basedir, winstate.snapcnt);
    printf("dumping image to file '%s'\n", filename);
    winstate.snapcnt++;

    fp = fopen(filename, "wb");		// write binary mode
    if (fp == NULL) {
	printf("failed trying to create snapshot file '%s'\n", filename);
	return;
    }

    // write headers
    (void)fwrite(&bmfh,    sizeof(bmfh),    1, fp);
    (void)fwrite(&bmih,    sizeof(bmih),    1, fp);
    (void)fwrite(&aColors, sizeof(RGBQUAD), 2, fp);

    // write bitmap -- 1bpp dib can apparently be only bottom-up
    // this code assumes fwrite() always writes the full amount.
    ip = imgmap;
    ep = imgmap + imgbytes;
    for(r=winstate.scrn_h-1; r>=0; r--) {
	(void)fwrite(imgmap + winstate.scrn_stride*r,
		     sizeof(byte), winstate.scrn_stride, fp);
    }

    // we're done -- clean up
    fclose(fp);
    free(imgmap);
}


// scan solace.ini for the options.
static void
LoadOptions(void)
{
    char ini_file[_MAX_PATH];
    char buf[80];
    int top, bottom, left, right;
    int i;
    UINT arg;
    DWORD fstat;

    strcpy(ini_file, winstate.basedir);
    strcat(ini_file, g_ini_file);

    fstat = GetFileAttributes(ini_file);
    if (fstat == 0xFFFFFFFF)
	return;

    GetPrivateProfileString("cpu", "speed", "2.04MHz", buf, sizeof(buf), ini_file);
    if (!strcmp(buf, "999MHz"))
	simstate.up_speed = MHZ_unregulated;
    else if (!strcmp(buf, "2.86MHz"))
	simstate.up_speed = MHZ_2_86;
    else if (!strcmp(buf, "2.38MHz"))
	simstate.up_speed = MHZ_2_38;
    else
	simstate.up_speed = MHZ_2_04;

    GetPrivateProfileString("display", "color", "white", buf, sizeof(buf), ini_file);
    if (!strcmp(buf, "amber"))
	winstate.disp_color = DISP_AMBER;
    else if (!strcmp(buf, "green"))
	winstate.disp_color = DISP_GREEN;
    else
	winstate.disp_color = DISP_WHITE;

    winstate.hidemouse = GetPrivateProfileInt("display", "hidemouse", 0, ini_file);
    winstate.stats     = GetPrivateProfileInt("display", "showstats", 0, ini_file);

    // state for four dip switches
    for(i=1; i<5; i++) {
	char buf2[10], buf3[10];
	sprintf(buf2, "sw%d", i);
	sprintf(buf3, "0x%02X", GetDipSwitch(i));
	GetPrivateProfileString("dipswitch", buf2, buf3,  buf, sizeof(buf), ini_file);
	sscanf(buf, "0x%02X", &arg);
	SetDipSwitch(i, (byte)arg);
    }

    // state for font selection method and current font
    GetPrivateProfileString("charset", "mode", "normal", buf, sizeof(buf), ini_file);
    if (!strcmp(buf, "dual"))
	winstate.fontmode = mainnow;	// allow toggle between main and alt
    else
	winstate.fontmode = mainonly;	// only main
    GetPrivateProfileString("charset", "mainfont", "6574", buf, sizeof(buf), ini_file);
    winstate.pMainFont = (!strcmp(buf, "6575")) ? charset_6575
						: charset_6574;
    GetPrivateProfileString("charset", "altfont",  "Devanagri", buf, sizeof(buf), ini_file);
    winstate.pAltFont = (!strcmp(buf, "Dzonkha")) ? charset_dzonkha
						  : charset_devanagri;
    winstate.pFont = (winstate.fontmode == altnow) ? winstate.pAltFont
						   : winstate.pMainFont;

    // restore window size and placement
    GetPrivateProfileString("windows", "main", "", buf, sizeof(buf), ini_file);
    if (sscanf(buf, "%d,%d,%d,%d", &top, &left, &bottom, &right) == 4) {
#if 1
	MoveWindow(winstate.hWnd, left, top, right-left, bottom-top, FALSE);
#else
	SetWindowPos(winstate.hWnd, HWND_TOP,
		     left, top, right-left, bottom-top,
		     SWP_NOACTIVATE);
#endif
    }

    // save default directories for ROMs and programs
    GetPrivateProfileString("files", "ROMname", "", buf, sizeof(buf), ini_file);
    if (strlen(buf) > 0) {
	char *path = extract_path(buf);
	strcpy((char*)winstate.ofn[0].lpstrFile, buf);
	strcpy((char*)winstate.ofn[0].lpstrInitialDir, path);
	free(path);
    }
    GetPrivateProfileString("files", "PROGdir", "", buf, sizeof(buf), ini_file);
    if (strlen(buf) > 0)
	strcpy((char*)winstate.ofn[1].lpstrInitialDir, buf);

    // get options for tape interface
    WinTapeLoadOptions(ini_file);
}


// save program options to solace.ini.
// the ini file is located wherever the current program is located.
static void
SaveOptions(void)
{
    char ini_file[_MAX_PATH];
    char buf[80];
    char *str;
    RECT rc;
    DWORD fstat;

    strcpy(ini_file, winstate.basedir);
    strcat(ini_file, g_ini_file);

    fstat = GetFileAttributes(ini_file);
    if (fstat == 0xFFFFFFFF) {
	int rslt;
	sprintf(buf, "Creating file '%s'", ini_file);
	rslt = MessageBox(winstate.hWnd, buf, "Confirm",
			    MB_OKCANCEL | MB_ICONEXCLAMATION | MB_TOPMOST);
	if (rslt != IDOK)
	    return;
    } else if (fstat & (FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_DIRECTORY)) {
	sprintf(buf, "Can't open file '%s': it is not writable", ini_file);
	WinAlert(buf);
	return;
    }

    switch (simstate.up_speed) {
	case MHZ_2_04: str = "2.04MHz"; break;
	case MHZ_2_38: str = "2.38MHz"; break;
	case MHZ_2_86: str = "2.86MHz"; break;
	      default: str = "999MHz";  break;
    }
    WritePrivateProfileString("cpu", "speed", str, ini_file);

    switch (winstate.disp_color) {
	case DISP_WHITE: str = "white"; break;
	case DISP_GREEN: str = "green"; break;
	case DISP_AMBER: str = "amber"; break;
    }
    WritePrivateProfileString("display", "color", str, ini_file);

    str = (winstate.hidemouse) ? "1" : "0";
    WritePrivateProfileString("display", "hidemouse", str, ini_file);
    str = (winstate.stats) ? "1" : "0";
    WritePrivateProfileString("display", "showstats", str, ini_file);

    sprintf(buf, "0x%02X", GetDipSwitch(1));
    WritePrivateProfileString("dipswitch", "sw1", buf, ini_file);
    sprintf(buf, "0x%02X", GetDipSwitch(2));
    WritePrivateProfileString("dipswitch", "sw2", buf, ini_file);
    sprintf(buf, "0x%02X", GetDipSwitch(3));
    WritePrivateProfileString("dipswitch", "sw3", buf, ini_file);
    sprintf(buf, "0x%02X", GetDipSwitch(4));
    WritePrivateProfileString("dipswitch", "sw4", buf, ini_file);

    // it doesn't seem likely that the original hardware could
    // remember the state of the font chosen between power cycles,
    // so all we keep is whether we allow an alternate mode.
    str = (winstate.fontmode == mainonly) ? "normal" : "dual";
    WritePrivateProfileString("charset", "mode", str, ini_file);
    str = MainFontName();
    WritePrivateProfileString("charset", "mainfont", str, ini_file);
    str = AltFontName();
    WritePrivateProfileString("charset", "altfont", str, ini_file);

    // save window size and placement
    GetWindowRect(winstate.hWnd, &rc);
    sprintf(buf, "%d,%d,%d,%d", rc.top, rc.left, rc.bottom, rc.right);
    WritePrivateProfileString("windows", "main", buf, ini_file);

    // save default directories for ROMs and programs
    if (strlen(winstate.ofn[0].lpstrFile) > 0)
	WritePrivateProfileString("files", "ROMname", winstate.ofn[0].lpstrFile, ini_file);
    if (strlen(winstate.ofn[1].lpstrInitialDir) > 0)
	WritePrivateProfileString("files", "PROGdir", winstate.ofn[1].lpstrInitialDir, ini_file);

    // save options for tape interface
    WinTapeSaveOptions(ini_file);
}


// -------------- manipulate status bar ------------------

// build a status bar, located at the bottom of the window
static void
BuildStatusBar(HWND hwnd)
{
    static int iStatusWidths[] = {300, -1};
    RECT rStatus;

    InitCommonControls();

    // create a basic status bar
    winstate.hStatus = CreateStatusWindow(WS_CHILD | WS_VISIBLE,
				    (LPCTSTR)NULL, hwnd, StatusID);

    // tell the status bar it has two message areas
    SendMessage(winstate.hStatus, SB_SIMPLE, (WPARAM)FALSE, (LPARAM)0);
    SendMessage(winstate.hStatus, SB_SETPARTS, 2, (LPARAM)iStatusWidths);

    // save its height for building entire client window geometry
    GetWindowRect(winstate.hStatus, &rStatus);
    winstate.status_h = rStatus.bottom - rStatus.top;
}


void
StatusBarMsg(int subwin, char *msg)
{
    // we report stats only when user wants them
    if (subwin == 1 && !winstate.stats)
	return;

    SendMessage(winstate.hStatus, SB_SETTEXT, subwin, (LPARAM)msg);
}


static char *
MainFontName(void)
{ return (winstate.pMainFont == charset_6574) ? "6574" : "6575"; }

static char *
AltFontName(void)
{ return (winstate.pAltFont == charset_dzonkha) ? "Dzonkha" : "Devanagri"; }


// return a pointer to an allocated string containing the path leading
// to the absolute file name passed in.  the caller must free the
// returned string at some point.
char *
extract_path(char *name)
{
    char *path, *p;
    int len;

    ASSERT(name != NULL);

    len = strlen(name);
    ASSERT(strlen(name) > 0);

    path = (char*)malloc(strlen(name) + 1);
    ASSERT(path != NULL);

    // copy the current filename into the initial dir...
    strcpy(path, (const char*)name);

    // ...then erase the text after the last backslash.
    p = &path[len - 1];
    while (*p != '\\') {
	ASSERT(p > path); // at least one backslash
	p--;
    }
    *p = '\0';	// wipe out tail

    return path;
}


// returns 1 if the filename passed in is absolute
int
absolute_path(char *name)
{
    ASSERT(name != NULL);

    if (name[0] == '\\')	// EG: "\emul\tapes" or "\\server\foo\bar"
	return 1;
    if (name[1] == ':')		// EG: "c:\emul\tapes"
	return 1;
    // does that cover it for windows?

    return 0;
}


// -------------- system timer functions ------------------

// figure out what type of timing capability the machine
// supports, and return the frequency of the time reference.
// we save the timer capabilities so CurTick() knows which
// facility to use.

uint32
QueryTimebase(void)
{
    LARGE_INTEGER HPTimerFreq;
    DWORD freq;

    winstate.useHPTimer = QueryPerformanceFrequency(&HPTimerFreq);
#if 0
    winstate.useHPTimer = FALSE; // for testing purposes
#endif
    winstate.useLPTimer = FALSE;

    if (winstate.useHPTimer) {

	printf("This machine has a high-performance timer of %I64d Hz\n", HPTimerFreq);
	ASSERT(HPTimerFreq.HighPart == 0);
	freq = HPTimerFreq.LowPart;

    } else {

	MMRESULT mm;
	printf("Using 1ms timer function\n");
	mm = timeBeginPeriod(1);	// 1 ms timer resolution
	winstate.useLPTimer = (mm == TIMERR_NOERROR);
	mm = timeEndPeriod(1);	// must be paired with timeBeginPeriod()
	if (!winstate.useLPTimer) {
	    printf("Hey, mm timer can't support 1ms timer resolution\n");
	    freq = 0;		// no useful timer
	}
	freq = 1000;		// 1 KHz timer resolution
    }

    winstate.tick_freq = freq;	// useful for debugging

#if 0
    // for testing purposes
    winstate.useHPTimer = FALSE;
    winstate.useLPTimer = FALSE;
    freq = 0;
#endif

    return freq;
}


// return a 32b current time count, taking care of available timer calls.
uint32
CurTick(void)
{
    if (winstate.useHPTimer) {
	LARGE_INTEGER time;
	(void)QueryPerformanceCounter(&time);
	return time.LowPart;
    } else if (winstate.useLPTimer) {
	DWORD time;
	(void)timeBeginPeriod(1);	// 1 ms timer resolution
	time = timeGetTime();
	(void)timeEndPeriod(1);		// must be paired with timeBeginPeriod()
	return time;
    }

    return 0;
}


uint32
ElapsedTick(uint32 begin, uint32 end)
{
    // I think wrapping takes care of itself because everything is unsigned
    uint32 now = CurTick();
    return now - begin;
}

