// Solace -- Sol Anachronistic Computer Emulation
// A Win32 emulator for the Sol-20 computer.
//
// Copyright (c) Jim Battle, 2000, 2001

// Create a modeless dialog box with five tabbed sheets,
// one for each of the four dipswitches on the board,
// plus one for the character generator options.

#include <stdlib.h>
#include <stdio.h>
#include <windows.h>
#include <windowsx.h>

#include "solace.h"
#include "resource.h"


#define numtabs 6

// the .repositioned member below is a bit ugly, but I couldn't
// find a message or callback that triggered at the right time
// at the propsheet creation.  instead, I reposition the propsheet
// at the time the first page (whichever that may be) is initialized,
// which happens before the window gets shown.  If I simply try to
// call SetWindowPlacement() after the call to PropertySheet(),
// it causes the propsheet to pop up in some default location then
// move to its new location.  much ado about nothing.

static struct {
    PROPSHEETHEADER pshdr;		// prop sheet header structure
    PROPSHEETPAGE   pspage[numtabs];	// prop sheet page structures
    BOOL            placement_valid;	// is debugger window position set?
    BOOL            repositioned;	// have we set position of prop sheet yet?
    WINDOWPLACEMENT placement;		// size, position, open/close state of dialog
    HWND            hwndPage[numtabs];	// handle to page
} optstate;


// called once at startup
void
InitOpts(void)
{
    int i;

    optstate.placement_valid = FALSE;
    optstate.repositioned    = FALSE;

    for(i=0; i<numtabs; i++)
	optstate.hwndPage[i] = NULL;
}


static void
Refresh_dips(HWND hwndDlg, int byte)
{
    static int ids[] = { IDC_DIP_1, IDC_DIP_2, IDC_DIP_3, IDC_DIP_4,
                         IDC_DIP_5, IDC_DIP_6, IDC_DIP_7, IDC_DIP_8 };
    int i;
    for(i=0; i<8; i++)
	CheckDlgButton(hwndDlg, ids[i], (byte>>i) & 1);
}


static int
MapSW2Int(WPARAM wParam)
{
    int n;
    switch (wParam) {
	case IDC_DIP_1: n=0; break;
	case IDC_DIP_2: n=1; break;
	case IDC_DIP_3: n=2; break;
	case IDC_DIP_4: n=3; break;
	case IDC_DIP_5: n=4; break;
	case IDC_DIP_6: n=5; break;
	case IDC_DIP_7: n=6; break;
	case IDC_DIP_8: n=7; break;
    }
    return n;
}


// the Sol hardware design is cheesy in certain areas.  For instance,
// there are user accessible dip switches whereby some outputs get
// shorted together and fight, or worse, as in this case, an output
// gets shorted to ground, potentially damaging a chip.
static void
SW1_Error(HWND hWnd)
{
    (void)MessageBox(hWnd,
	    "You can't have both SW1-5 and SW1-6 on at the same time.\n" \
	    "You've blown the output driver of U88, pin 6.\n" \
	    "Time to break out the soldering iron!",
	    "Bozo Box",
	    MB_OK | MB_ICONEXCLAMATION);
}


// this function is called on creation/destruction of prop sheet
// it is used to associate a switch number with a given window.
//
// this is pretty useless ... hwnd is null, and ppsp doesn't
// provide a way to recover the window handle for the sheet
// it is associated with, so instead we resort to maintaining
// it in a global array; first one here deletes them all.

static UINT CALLBACK
DipSheetProc(HWND hwnd, UINT uMsg, LPPROPSHEETPAGE ppsp)
{
    int i;

    switch (uMsg) {

	case PSPCB_CREATE:
	    return TRUE;

	case PSPCB_RELEASE:
	    for(i=0; i<4; i++)
		if (optstate.hwndPage[i] != NULL) {
		    RemoveProp(optstate.hwndPage[i], "switchnum");
		    optstate.hwndPage[i] = NULL;
		}
	    return 0;
    }

    return 0;
}


static void
SaveOptPosition(void)
{
    if (winstate.hOpts != NULL) {
	optstate.placement.length = sizeof(WINDOWPLACEMENT);
	GetWindowPlacement(winstate.hOpts, &optstate.placement);
	optstate.placement_valid = TRUE;
	optstate.repositioned    = FALSE;
	DestroyWindow(winstate.hOpts);
	winstate.hOpts = NULL;
    }
}


static BOOL
handle_sw1(HWND hwndDlg, byte dip, WPARAM wParam, int n)
{
    byte newdip = (byte)(dip ^ (1<<n));

    switch (wParam) {

	case IDC_DIP_2:
	case IDC_DIP_7:
	case IDC_DIP_8:
	    SetDipSwitch(1, newdip);
	    return TRUE;

	case IDC_DIP_1:
	    SetDipSwitch(1, newdip);
	    WinResetSim(TRUE);
	    return TRUE;

	case IDC_DIP_3:
	    SetDipSwitch(1, newdip);
	    update_disp(0,0,1);
	    return TRUE;

	case IDC_DIP_4:
	    SetDipSwitch(1, newdip);
	    SetDispColors();
	    InvalidateRect(hwndDlg, NULL, TRUE);
	    return TRUE;

	case IDC_DIP_5:
	    if (  (dip & SW1_SOLIDCUR) &&
		 !(dip & SW1_BLINKCUR)) {
		SW1_Error(hwndDlg);
		Refresh_dips(hwndDlg, dip);
	    } else {
		SetDipSwitch(1, newdip);
		update_disp(0,1,0);
	    }
	    return TRUE;

	case IDC_DIP_6:
	    if ( !(dip & SW1_SOLIDCUR) &&
		  (dip & SW1_BLINKCUR)) {
		SW1_Error(hwndDlg);
		Refresh_dips(hwndDlg, dip);
	    } else {
		SetDipSwitch(1, newdip);
		update_disp(0,1,0);
	    }
	    return TRUE;
    } // WM_COMMAND switch

    return FALSE;	// shouldn't ever happen
}


static BOOL
handle_sw3(HWND hwndDlg, byte dip, WPARAM wParam, int n)
{
    byte newdip = dip ^ (1<<n);
    if ((newdip & (newdip-1)) != 0) {
	(void)MessageBox(hwndDlg,
		"Only one of SW3-1 to SW3-8 can be on at the same time.\n" \
		"You've herniated the output driver of U86.\n" \
		"Time to break out the soldering iron!",
		"Bozo Box",
		MB_OK | MB_ICONEXCLAMATION);
	Refresh_dips(hwndDlg, dip);
    } else
	SetDipSwitch(3, newdip);
    return TRUE;
}


static BOOL CALLBACK
SWDlgProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    int switchnum;
    LPNMHDR pnmh;
    byte dip;
    int n;

    if (msg != WM_INITDIALOG) {
	// retrieve the switch number associated with the window handle
	switchnum = (int)GetProp(hwndDlg, "switchnum");
	if (switchnum > 0)
	    dip = GetDipSwitch(switchnum);
    }

    switch (msg) {

	case WM_INITDIALOG:
	    // associate the switch number with the window handle
	    n = (int)(((PROPSHEETPAGE*)lParam)->lParam);
	    SetProp(hwndDlg, "switchnum", (HANDLE)n);
	    optstate.hwndPage[n] = hwndDlg;
	    if (optstate.placement_valid && !optstate.repositioned) {
		SetWindowPlacement(GetParent(hwndDlg), &optstate.placement);
		optstate.repositioned = TRUE;
	    }
	    return TRUE;

	case WM_COMMAND:
	    n = MapSW2Int(wParam);
	    switch (switchnum) {
		case 1:
		    return handle_sw1(hwndDlg, dip, wParam, n);
		case 3:
		    return handle_sw3(hwndDlg, dip, wParam, n);
		case 2:
		case 4:
		    SetDipSwitch(switchnum, (byte)(dip ^ (1<<n)));
		    return TRUE;
	    }
	    return FALSE;


	case WM_NOTIFY:
	    pnmh = (LPNMHDR)lParam;
	    switch (pnmh->code) {

		case PSN_SETACTIVE:
		    Refresh_dips(hwndDlg, dip);
		    return TRUE;

		case PSN_QUERYCANCEL:
		case PSN_APPLY:
		    // closing dialog
		    SaveOptPosition();
		    return TRUE;
	    }
	    break;

	default:
	    break;
    }

    return FALSE;	// didn't handle the message
}


static void
set_fontmode_display(HWND hwndDlg)
{
    HWND hMainFont = GetDlgItem(hwndDlg, IDC_CGEN_BASEFONT);
    HWND hAltFont  = GetDlgItem(hwndDlg, IDC_CGEN_ALTFONT);
    HWND hFontMode = GetDlgItem(hwndDlg, IDC_CGEN_DH);
    HWND hText     = GetDlgItem(hwndDlg, IDC_STATIC_ALT);
    BOOL b = (winstate.fontmode != mainonly);

    ComboBox_SetCurSel(hMainFont, (winstate.pMainFont == charset_6575));
    ComboBox_SetCurSel(hAltFont,  (winstate.pAltFont  == charset_dzonkha));

    Button_SetCheck(hFontMode, b ? BST_CHECKED : BST_UNCHECKED);
    EnableWindow(hAltFont, b);
    EnableWindow(GetDlgItem(hwndDlg, IDC_STATIC_ALT), b);
}


static BOOL CALLBACK
CharsetDlgProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    HWND hMainFont = GetDlgItem(hwndDlg, IDC_CGEN_BASEFONT);
    HWND hAltFont  = GetDlgItem(hwndDlg, IDC_CGEN_ALTFONT);
    HWND hFontMode = GetDlgItem(hwndDlg, IDC_CGEN_DH);
    LPNMHDR pnmh;
    int n;

    switch (msg) {

	case WM_INITDIALOG:
	    ComboBox_InsertString(hMainFont, 0, (LPARAM)"6574");
	    ComboBox_InsertString(hMainFont, 1, (LPARAM)"6575");
	    ComboBox_InsertString(hAltFont,  0, (LPARAM)"Devanagri");
	    ComboBox_InsertString(hAltFont,  1, (LPARAM)"Dzonkha");
	    set_fontmode_display(hwndDlg);
	    return TRUE;

	case WM_COMMAND:
	    switch (LOWORD(wParam)) {

		case IDC_CGEN_DH:
		    if (HIWORD(wParam) == BN_CLICKED) {
			int bst = Button_GetState(hFontMode) & 0x0003;
			winstate.fontmode = (bst == BST_CHECKED) ? mainnow : mainonly;
			CreateCharMap();
			set_fontmode_display(hwndDlg);
			return TRUE;
		    }
		    return FALSE;

		case IDC_CGEN_BASEFONT:
		    switch (HIWORD(wParam)) {
			case CBN_SELENDOK:
			    n = ComboBox_GetCurSel(hMainFont);
			    winstate.pMainFont = (n) ? charset_6575
						     : charset_6574;
			    set_fontmode_display(hwndDlg);
			    CreateCharMap();
			    return TRUE;
			default:
			    break;
		    }
		    return FALSE;

		case IDC_CGEN_ALTFONT:
		    switch (HIWORD(wParam)) {
			case CBN_SELENDOK:
			    n = ComboBox_GetCurSel(hAltFont);
			    winstate.pAltFont = (n) ? charset_dzonkha
						    : charset_devanagri;
			    set_fontmode_display(hwndDlg);
			    CreateCharMap();
			    return TRUE;
			default:
			    break;
		    }
		    return FALSE;

	    }
	    return FALSE;	// shouldn't ever happen


	case WM_NOTIFY:
	    pnmh = (LPNMHDR)lParam;
	    switch (pnmh->code) {

		case PSN_SETACTIVE:
		    return TRUE;

		case PSN_QUERYCANCEL:
		case PSN_APPLY:
		    // closing dialog
		    SaveOptPosition();
		    return TRUE;
	    }
	    break;


	default:
	    break;
    }

    return FALSE;	// didn't handle the message
}


static void
set_svtopts_display(HWND hwndDlg)
{
    HWND hReal = GetDlgItem(hwndDlg, IDC_SVT_SPEEDREAL);
    HWND hFast = GetDlgItem(hwndDlg, IDC_SVT_SPEEDFAST);
    HWND hBaud = GetDlgItem(hwndDlg, IDC_SVT_IGNOREBAUD);
    int realtime, ignorebaud;

    GetTapeProp(1, TPROP_REALTIME,   &realtime);
    GetTapeProp(1, TPROP_IGNOREBAUD, &ignorebaud);

    Button_SetCheck(hReal,  realtime   ? BST_CHECKED : BST_UNCHECKED);
    Button_SetCheck(hFast, !realtime   ? BST_CHECKED : BST_UNCHECKED);
    Button_SetCheck(hBaud,  ignorebaud ? BST_CHECKED : BST_UNCHECKED);
}


static BOOL CALLBACK
SVTDlgProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    LPNMHDR pnmh;

    switch (msg) {

	case WM_INITDIALOG:
	    set_svtopts_display(hwndDlg);
	    return TRUE;

	case WM_COMMAND:
	    switch (LOWORD(wParam)) {

		case IDC_SVT_SPEEDREAL:
		    if (HIWORD(wParam) == BN_CLICKED) {
			SetTapeProp(0, TPROP_REALTIME, 1);
			return TRUE;
		    }
		    return FALSE;

		case IDC_SVT_SPEEDFAST:
		    if (HIWORD(wParam) == BN_CLICKED) {
			SetTapeProp(0, TPROP_REALTIME, 0);
			return TRUE;
		    }
		    return FALSE;

		case IDC_SVT_IGNOREBAUD:
		    if (HIWORD(wParam) == BN_CLICKED) {
			int bst = Button_GetState(GetDlgItem(hwndDlg, IDC_SVT_IGNOREBAUD)) & 0x0003;
			SetTapeProp(0, TPROP_IGNOREBAUD, bst);
			SetTapeProp(1, TPROP_IGNOREBAUD, bst);
			return TRUE;
		    }
		    return FALSE;

	    }
	    return FALSE;	// shouldn't ever happen


	case WM_NOTIFY:
	    pnmh = (LPNMHDR)lParam;
	    switch (pnmh->code) {

		case PSN_SETACTIVE:
		    return TRUE;

		case PSN_QUERYCANCEL:
		case PSN_APPLY:
		    // closing dialog
		    SaveOptPosition();
		    return TRUE;
	    }
	    break;


	default:
	    break;
    }

    return FALSE;	// didn't handle the message
}


// Propery sheets have an "OK" and a "CANCEL" button.  We don't want to
// have a cancel button at all, but if we just remove the cancel button,
// the OK button is in the wrong place.  So we go throught the steps below.
static void
HideCancelButton(void)
{
    HWND hwndCancel = GetDlgItem(winstate.hOpts, IDCANCEL);
    HWND hwndOK     = GetDlgItem(winstate.hOpts, IDOK);
#if 0
// the following code works fine on Win95/98, but for some
// unknown reason causes a core dump on NT.  More specifically,
// *this* code doesn't; even if this routine isn't called,
// clicking on the OK button results in a core dump, while
// clicking on CANCEL doesn't, even though the two cases take
// the same code path in the solace user code.
    RECT rc;

    // find location of CANCEL button
    GetWindowRect(hwndCancel, &rc);
    ScreenToClient(winstate.hOpts, (LPPOINT)&rc.left);
    ScreenToClient(winstate.hOpts, (LPPOINT)&rc.right);

    // disable cancel button
    EnableWindow(hwndCancel, FALSE);
    ShowWindow(hwndCancel, FALSE);

    // change label on OK button
    SetWindowText(hwndOK, "Done");

    // move OK button to where Cancel was
    (void)MoveWindow(hwndOK,
	  rc.left, rc.top, rc.right-rc.left, rc.bottom-rc.top,
	  TRUE);
#else
// because clicking on DONE causes a core dump on NT 4 (s.p. 3)
// for me, here we take a different approach.  We disable the
// OK button and change the label on CANCEL to read DONE instead.
// it is actually less code, but more obscure.

    // disable the OK button
    EnableWindow(hwndOK, FALSE);
    ShowWindow(hwndOK, FALSE);

    // change label on the Cancel button
    SetWindowText(hwndCancel, "Done");
#endif
}


void
CreateOptsPropSheet(void)
{
    int i;

    ZeroMemory(&optstate.pshdr, sizeof(PROPSHEETHEADER));
    ZeroMemory(&optstate.pspage, numtabs*sizeof(PROPSHEETPAGE));

    optstate.pshdr.dwSize  = sizeof(PROPSHEETHEADER);
    optstate.pshdr.dwFlags = PSH_MODELESS	// modeless dialog
			   | PSH_PROPSHEETPAGE	// use psp structures
			   | PSH_NOAPPLYNOW;	// hide APPLY button
    optstate.pshdr.hInstance   = winstate.hInst;
    optstate.pshdr.pszCaption  = "Emulator Options";
    optstate.pshdr.nPages      = numtabs;
    optstate.pshdr.nStartPage  = 0;
    optstate.pshdr.ppsp        = optstate.pspage;
    optstate.pshdr.pfnCallback = NULL;	// only if PSH_USECALLBACK

    for(i=0; i<4; i++) {
	static WORD template[] =
		{ IDD_DIPSW1, IDD_DIPSW2, IDD_DIPSW3, IDD_DIPSW4 };
	optstate.pspage[i].dwSize      = sizeof(PROPSHEETPAGE);
	optstate.pspage[i].dwFlags     = PSP_USECALLBACK;
	optstate.pspage[i].hInstance   = winstate.hInst;
	optstate.pspage[i].pszTemplate = MAKEINTRESOURCE(template[i]);
	optstate.pspage[i].pfnDlgProc  = SWDlgProc;
	optstate.pspage[i].lParam      = (LPARAM)(i+1);
	optstate.pspage[i].pfnCallback = DipSheetProc;
    }

    optstate.pspage[4].dwSize      = sizeof(PROPSHEETPAGE);
    optstate.pspage[4].dwFlags     = 0;
    optstate.pspage[4].hInstance   = winstate.hInst;
    optstate.pspage[4].pszTemplate = MAKEINTRESOURCE(IDD_CHARSET);
    optstate.pspage[4].pfnDlgProc  = CharsetDlgProc;
    optstate.pspage[4].pfnCallback = NULL;

    optstate.pspage[5].dwSize      = sizeof(PROPSHEETPAGE);
    optstate.pspage[5].dwFlags     = 0;
    optstate.pspage[5].hInstance   = winstate.hInst;
    optstate.pspage[5].pszTemplate = MAKEINTRESOURCE(IDD_SVT);
    optstate.pspage[5].pfnDlgProc  = SVTDlgProc;
    optstate.pspage[5].pfnCallback = NULL;

    // OK, create it!
    winstate.hOpts = (HWND)PropertySheet(&optstate.pshdr);

    HideCancelButton();
}

